// MPIJob.cpp : Implementation of CMPIJob
#include "stdafx.h"
#include "MPIJobServer.h"
#include "MPIJob.h"
#include "..\Common\MPIJobDefs.h"
#include "..\Common\Translate_Error.h"
#include "..\Common\MPIJobLog.h"
// You must compile MPILaunchServer first because the following files are generated
#include "..\MPILaunchServer\MPILaunchServer.h"
#include "..\MPILaunchServer\MPILaunchserver_i.c"
#include <stdio.h>
#include <time.h>

/////////////////////////////////////////////////////////////////////////////
// CMPIJob

// Function name	: CMPIJob::~CMPIJob
// Description	    : 
// Return type		: 
CMPIJob::~CMPIJob()
{
	long error;
	BSTR berror_msg = SysAllocString(L"");
	DestroyJob(&error, &berror_msg);
	SysFreeString(berror_msg);

	CloseHandle(m_hOutputEvent);
	CloseHandle(m_hOutputMutex);
	CloseHandle(m_hInputEvent);
	CloseHandle(m_hInputMutex);
	CloseHandle(m_hAbortEvent);
	if (m_hTempFile != NULL)
	{
		CloseHandle(m_hTempFile);
		m_hTempFile = NULL;
	}
	if (m_pOutList)
	{
		OutputNode *n;

		while (m_pOutList != NULL)
		{
			n = m_pOutList;
			m_pOutList = m_pOutList->next;
			if (n->data != NULL)
				delete n->data;
			delete n;
		}
	}
	m_pOutList = NULL;
	m_pOutListTail = NULL;
	if (m_pInList)
	{
		InputNode *n;

		while (m_pInList != NULL)
		{
			n = m_pInList;
			m_pInList = m_pInList->next;
			if (n->data != NULL)
				delete n->data;
			delete n;
		}
	}
	m_pInList = NULL;
	m_pInListTail = NULL;
}

// Function name	: Connect
// Description	    : 
// Return type		: HRESULT 
// Argument         : WCHAR *host
// Argument         : IMPIJob *&pJob
HRESULT Connect(WCHAR *host, IMPIJob *&pJob)
{
	HRESULT hr;
	MULTI_QI qi;
	COSERVERINFO server;
	WCHAR localhost[100];
	DWORD length = 100;
	WCHAR error_msg[256];

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	if (FAILED(hr))
	{
		LogMsg(TEXT("CoInitializeEx failed\n"));
		return hr;
	}

	qi.pIID = &IID_IMPIJob;
	qi.pItf = NULL;

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;
	server.pwszName = host;

	GetComputerNameW(localhost, &length);

	if (_wcsicmp(localhost, host) == 0)
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			Translate_HRError(hr, error_msg);
			LogWMsg(L"Unable to connect to %s: %s\n", host, error_msg);
			return hr;
		}
	}
	else
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_REMOTE_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			Translate_HRError(hr, error_msg);
			LogWMsg(L"Unable to connect to %s: %s\n", host, error_msg);
			return hr;
		}
	}

	if (SUCCEEDED(qi.hr))
	{
		pJob = (IMPIJob*)qi.pItf;
		return S_OK;
	}

	return qi.hr;
}

// Function name	: Connect
// Description	    : 
// Return type		: HRESULT 
// Argument         : IMPILaunch *&pLaunch
HRESULT Connect(IMPILaunch *&pLaunch)
{
	HRESULT hr;
	MULTI_QI qi;
	COSERVERINFO server;
	WCHAR localhost[100];
	DWORD length = 100;
	WCHAR error_msg[256];

	qi.pIID = &IID_IMPILaunch;
	qi.pItf = NULL;

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;

	GetComputerNameW(localhost, &length);
	server.pwszName = localhost;

	hr = CoCreateInstanceEx(CLSID_MPILaunch, NULL, CLSCTX_SERVER, &server, 1, &qi);
	if (FAILED(hr))
	{
		Translate_HRError(hr, error_msg);
		LogWMsg(L"Unable to connect to %s: %s\n", localhost, error_msg);
		return hr;
	}

	if (SUCCEEDED(qi.hr))
	{
		pLaunch = (IMPILaunch*)qi.pItf;
		return S_OK;
	}

	return qi.hr;
}

// Function name	: ExtendJobCollectOutputThread
// Description	    : 
// Return type		: void 
// Argument         : CollectOutputThreadArg *arg
void ExtendJobCollectOutputThread(CollectOutputThreadArg *arg)
{
	IMPIJob *pJob=NULL;
	CMPIJob *pLocalJob;
	int nIproc;
	HRESULT hr=S_OK;
	long error=0;
	BSTR berror_msg;
	VARIANT *v;
	long more = 1, from = 0;
	SAFEARRAYBOUND bound;
	DWORD num_elements=0;
	void *pBuf=NULL;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	pLocalJob = arg->pJob;
	nIproc = arg->nIproc;
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPIJob, (void**) &pJob);
	delete arg->ppStream;
	delete arg;
	arg = NULL;

	// Collect the output from the launched process
	berror_msg = SysAllocString(L"");
	v = new VARIANT;
	v->vt = VT_UI1 | VT_ARRAY;
	bound.lLbound = 0;
	bound.cElements = 0;
	v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
	
	DLogWMsg(L"ExtendJobCollectOutputThread: Getting data from extended job.\n");
	while (more == 1)
	{
		error = 0;
		hr = pJob->GetJobOutput(v, &from, &more, &error, &berror_msg);
		if (FAILED(hr))
		{
			WCHAR error_msg[256];
			Translate_HRError(hr, error_msg, L"ExtendJobCollectOutputThread:GetJobOutput failed ");
			LogWMsg(error_msg);
			VariantClear(v);
			delete v;
			pJob->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			return;
		}
		if (error)
		{
			LogWMsg(L"ExtendJobCollectOutputThread:GetJobOutput failed %d: %s", error, berror_msg);
			VariantClear(v);
			delete v;
			pJob->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			return;
		}
		if (v->parray != NULL)
		{
			num_elements = v->parray->rgsabound[0].cElements;
			if (num_elements > 0)
			{
				DLogWMsg(L"ExtendJobCollectOutputThread: %d bytes about to be inserted in list\n", num_elements);
				// Add the data to the output queue
				OutputNode *n = new OutputNode;
				n->data = new unsigned char[num_elements];
				SafeArrayAccessData(v->parray, &pBuf);
				memcpy(n->data, pBuf, num_elements);
				SafeArrayUnaccessData(v->parray);
				n->size = num_elements;
				n->from = from;
				n->next = NULL;
				if (WaitForSingleObject(pLocalJob->m_hOutputMutex, 10000) == WAIT_OBJECT_0)
				{
					if (pLocalJob->m_pOutListTail == NULL)
					{
						pLocalJob->m_pOutList = pLocalJob->m_pOutListTail = n;
					}
					else
					{
						pLocalJob->m_pOutListTail->next = n;
						pLocalJob->m_pOutListTail = n;
					}
					SetEvent(pLocalJob->m_hOutputEvent);
					ReleaseMutex(pLocalJob->m_hOutputMutex);
				}
				else
				{
					delete n->data;
					delete n;
				}

				// Destroy the array.
				SafeArrayDestroy(v->parray);
				// Create a new and empty array.
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
			}
			else
			{
				SafeArrayDestroy(v->parray);
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
			}
		}
		else
		{
			v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
		}
	}
	DLogWMsg(L"ExtendJobCollectOutputThread: no more data coming from YLaunchCOM.\n");
	
	VariantClear(v);
	delete v;
	SysFreeString(berror_msg);

	// Insert a termination node in the list
	OutputNode *n = new OutputNode;
	n->from = nIproc;
	n->size = 0; // <------ zero indicates the end of the stream
	n->data = NULL;
	n->next = NULL;
	if (WaitForSingleObject(pLocalJob->m_hOutputMutex, 10000) == WAIT_OBJECT_0)
	{
		if (pLocalJob->m_pOutListTail == NULL)
			pLocalJob->m_pOutList = pLocalJob->m_pOutListTail = n;
		else
		{
			pLocalJob->m_pOutListTail->next = n;
			pLocalJob->m_pOutListTail = n;
		}
		SetEvent(pLocalJob->m_hOutputEvent);
		ReleaseMutex(pLocalJob->m_hOutputMutex);
	}
	else
	{
		LogMsg(TEXT("Wait for OutputMutex failed in ExtendJobCollectOutput\n"));
		// Should I do this?  It is not a thread safe operation but how else am I going to signal that this stream has ended?
		//m_nOutputStreamsRemaining--;
		//InterlockedDecrement(m_nOutputStreamsRemaining);
		delete n;
	}

	pJob->Release();

	CoUninitialize();
}

// Function name	: ExtendJobThread
// Description	    : 
// Return type		: void 
// Argument         : ExtendJobThreadArg *arg
void ExtendJobThread(ExtendJobThreadArg *arg)
{
	HRESULT hr = S_OK;
	long error = 0;
	BSTR berror_msg;
	IMPIJob *pJob=NULL;
	WCHAR wNic[100];
	BSTR bNic, bJobID;
	DWORD size = 100;
	HANDLE hCollectOutputThread=NULL;

	try{
	DLogWMsg(L"Entering ExtendJobThread, job %s\n", arg->pJob->m_pszJobID);
	berror_msg = SysAllocString(L"");
	/*
	if (wcslen(arg->pJob->m_pszNic) > 0)
	{
		wcscpy(wNic, arg->pJob->m_pszNic);
	}
	else
	{
		if (!GetComputerNameW(wNic, &size))
		{
			TCHAR error_msg[256];
			arg->nResult = GetLastError();
			Translate_Error(arg->nResult, error_msg);
			LogMsg(TEXT("GetComputerName failed: error(%u) %s\n"), arg->nResult, error_msg);
			SetEvent(arg->hCommandCompletedEvent);
			SysFreeString(berror_msg);
			CoUninitialize();
			return;
		}
	}
	//*/
	wcscpy(wNic, arg->pszHost);

	clock_t start,finish;
	start = clock();
	hr = Connect(arg->pszHost, pJob);
	finish = clock();
	DLogMsg(TEXT("Connect took %.3f seconds\n"), (double)(finish-start)/(double)CLOCKS_PER_SEC);
	if (SUCCEEDED(hr))
	{
		bNic = SysAllocString(wNic);
		bJobID = SysAllocString(arg->pJob->m_pszJobID);
		if (arg->vHosts != NULL)
			hr = pJob->ExtendJob(arg->nIproc, arg->nNproc, bJobID, *(arg->vHosts), arg->nSMPProcs, bNic, &error, &berror_msg);
		else
		{
			VARIANT v;
			VariantInit(&v);
			hr = pJob->ExtendJob(arg->nIproc, arg->nNproc, bJobID, v, arg->nSMPProcs, bNic, &error, &berror_msg);
		}
		SysFreeString(bNic);
		SysFreeString(bJobID);
		if (FAILED(hr))
		{
			Translate_HRError(hr, arg->wErrorMsg, L"ExtendJobThread:ExtendJob failed ");
			LogWMsg(arg->wErrorMsg);
			arg->nResult = hr;
			SetEvent(arg->hCommandCompletedEvent);
			pJob->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			return;
		}
		if (error)
		{
			LogWMsg(L"ExtendJob failed in ExtendJobThread: %s\n", berror_msg);
			arg->nResult = error;
			swprintf(arg->wErrorMsg, L"ExtendJobThread:ExtendJob:%s", berror_msg);
			SetEvent(arg->hCommandCompletedEvent);
			pJob->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			return;
		}

		// Signal the thread is ready
		arg->nResult = S_OK;
		ResetEvent(arg->hCommandEvent);
		SetEvent(arg->hCommandCompletedEvent);

		while (true)
		{
			if (WaitForSingleObject(arg->hCommandEvent, INFINITE) != WAIT_OBJECT_0)
			{
				arg->nResult = GetLastError();
				Translate_Error(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:WaitForSingleObject(hCommandEvent) failed ");
				LogWMsg(arg->wErrorMsg);
				SetEvent(arg->hCommandCompletedEvent);
				pJob->Release();
				SysFreeString(berror_msg);
				CoUninitialize();
				return;
			}

			error = 0;
			switch(arg->nCommand)
			{
			case MPIJOB_THRD_CMD_EXIT:
				arg->nResult = 0;
				ResetEvent(arg->hCommandEvent);
				pJob->DestroyJob(&error, &berror_msg);
				pJob->Release();
				SysFreeString(berror_msg);
				CoUninitialize();
				DLogWMsg(L"Exiting ExtendJobThread, job %s\n", arg->pJob->m_pszJobID);
				SetEvent(arg->hCommandCompletedEvent);
				return;
				break;
			case MPIJOB_THRD_CMD_ABORT:
				ResetEvent(arg->hCommandEvent);
				arg->nResult = pJob->Abort(&error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					pJob->Release();
					SysFreeString(berror_msg);
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:Abort failed ");
					LogWMsg(arg->wErrorMsg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				if (error)
				{
					pJob->Release();
					swprintf(arg->wErrorMsg, L"ExtendJobThread:%s", berror_msg);
					SysFreeString(berror_msg);
					LogWMsg(arg->wErrorMsg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_SETEXE:
				ResetEvent(arg->hCommandEvent);
				error = 0;
				arg->nResult = pJob->SetExecutable((BSTR)arg->pCommandData, &error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					pJob->Release();
					SysFreeString(berror_msg);
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:SetExecutable failed ");
					wcscat(arg->wErrorMsg, (WCHAR*)arg->pCommandData);
					LogWMsg(L"SetExecutable failed: %s\n", arg->pCommandData);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				if (error)
				{
					pJob->Release();
					swprintf(arg->wErrorMsg, L"ExtendJobThread:(%s):%s", arg->pCommandData, berror_msg);
					LogWMsg(arg->wErrorMsg);
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_SETENV:
				ResetEvent(arg->hCommandEvent);
				error = 0;
				arg->nResult = pJob->SetEnvironmentVariables((BSTR)arg->pCommandData, &error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					pJob->Release();
					SysFreeString(berror_msg);
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:SetEnvironmentVariables failed ");
					wcscat(arg->wErrorMsg, (WCHAR*)arg->pCommandData);
					LogWMsg(L"SetEnvironmentVariables failed: %s\n", arg->pCommandData);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				if (error)
				{
					pJob->Release();
					swprintf(arg->wErrorMsg, L"ExtendJobThread:(%s):%s", arg->pCommandData, berror_msg);
					LogWMsg(arg->wErrorMsg);
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_SENDBREAK:
				ResetEvent(arg->hCommandEvent);
				error = 0;
				arg->nResult = pJob->SendBreak(&error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					pJob->Release();
					SysFreeString(berror_msg);
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:SendBreak failed ");
					wcscat(arg->wErrorMsg, (WCHAR*)arg->pCommandData);
					LogWMsg(L"SendBreak failed: %s\n", arg->pCommandData);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				if (error)
				{
					pJob->Release();
					swprintf(arg->wErrorMsg, L"ExtendJobThread:(%s):%s", arg->pCommandData, berror_msg);
					LogWMsg(arg->wErrorMsg);
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_SETDIR:
				ResetEvent(arg->hCommandEvent);
				arg->nResult = pJob->SetWorkingDirectory((BSTR)arg->pCommandData, &error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					pJob->Release();
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:SetWorkingDirectory failed ");
					wcscat(arg->wErrorMsg, (LPCWSTR)arg->pCommandData);
					LogWMsg(arg->wErrorMsg);
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				else
				if (error > 0)
				{
					// Don't fail if the working directory can't be set
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_STARTTRANSFER:
				ResetEvent(arg->hCommandEvent);
				arg->nResult = pJob->StartExeTransfer(&error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:StartExeTransfer failed ");
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				else
				if (error)
				{
					arg->nResult = error;
					swprintf(arg->wErrorMsg, L"ExtendJobThread:%s", berror_msg);
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_TRANSFER:
				ResetEvent(arg->hCommandEvent);
				error = 0;
				arg->nResult = pJob->TransferExeData( *((VARIANT*)arg->pCommandData), &error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:TransferExeData failed ");
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				else
				if (error)
				{
					arg->nResult = error;
					swprintf(arg->wErrorMsg, L"ExtendJobThread:%s", berror_msg);
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_FINISHTRANSFER:
				ResetEvent(arg->hCommandEvent);
				error = 0;
				arg->nResult = pJob->FinishExeTransfer(&error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:FinishExeTransfer failed ");
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				else
				if (error)
				{
					arg->nResult = error;
					swprintf(arg->wErrorMsg, L"ExtendJobThread:%s", berror_msg);
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				SetEvent(arg->hCommandCompletedEvent);
				break;
			case MPIJOB_THRD_CMD_LAUNCHJOB:
				ResetEvent(arg->hCommandEvent);
				arg->nResult = pJob->ExtendLaunchJob(
					((BSTR *)arg->pCommandData)[0], 
					((BSTR *)arg->pCommandData)[1], 
					((BSTR *)arg->pCommandData)[2], 
					((BSTR *)arg->pCommandData)[3], 
					((BSTR *)arg->pCommandData)[4], 
					arg->pJob->m_nLaunchFlags,
					&error, &berror_msg);
				if (FAILED(arg->nResult))
				{
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:ExtendLaunchJob failed ");
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				else
				if (error)
				{
					arg->nResult = error;
					swprintf(arg->wErrorMsg, L"ExtendJobThread:%s", berror_msg);
					LogWMsg(arg->wErrorMsg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}

				// marshall pJob to a thread which collects the job output
				CollectOutputThreadArg *pcArg = new CollectOutputThreadArg;
				pcArg->nIproc = arg->nIproc;
				pcArg->pJob = arg->pJob;
				pcArg->ppStream = new IStream*;
				arg->nResult = CoMarshalInterThreadInterfaceInStream(IID_IMPIJob, pJob, pcArg->ppStream);
				if (FAILED(arg->nResult))
				{
					Translate_HRError(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:CoMarshalInterThreadInterfaceInStream failed ");
					LogWMsg(arg->wErrorMsg);
					//pJob->Abort(&error, &berror_msg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}

				DWORD dwThreadID;
				hCollectOutputThread = CreateThread(
					NULL, 0, 
					(LPTHREAD_START_ROUTINE)ExtendJobCollectOutputThread,
					pcArg, 
					0, &dwThreadID);
				if (hCollectOutputThread == NULL)
				{
					arg->nResult = GetLastError();
					Translate_Error(arg->nResult, arg->wErrorMsg, L"ExtendJobThread:CreateThread(ExtendJobCollectOutputThread) failed ");
					LogWMsg(arg->wErrorMsg);
					//pJob->Abort(&error, &berror_msg);
					pJob->Release();
					SysFreeString(berror_msg);
					SetEvent(arg->hCommandCompletedEvent);
					CoUninitialize();
					return;
				}
				
				SetEvent(arg->hCommandCompletedEvent);
				break;
			}
		}

		pJob->Release();
		CoUninitialize();
	}
	else
	{
		swprintf(arg->wErrorMsg, L"ExtendJobThread:Connect to %s failed: ", arg->pszHost);
		Translate_HRError(hr, &arg->wErrorMsg[wcslen(arg->wErrorMsg)]);
		arg->nResult = hr;
		ResetEvent(arg->hCommandEvent);
		LogWMsg(arg->wErrorMsg);
		SetEvent(arg->hCommandCompletedEvent);
	}

	SysFreeString(berror_msg);
	//LogWMsg(L"Exiting ExtendJobThread, unknown how\n");
	}catch(...)
	{
		if (pJob != NULL)
			pJob->Release();
		arg->nResult = E_FAIL;
		SetEvent(arg->hCommandCompletedEvent);
		LogMsg(TEXT("ExtendJobThread:Exception thrown"));
	}
}

// Function name	: AbortThread
// Description	    : This thread simply waits on an event indefinitely.
//                    The event is set when the job terminates cleanly or Abort is called
// Return type		: void 
// Argument         : AbortThreadArg *arg
void AbortThread(AbortThreadArg *arg)
{
	IMPILaunch *pLaunch=NULL;
	HRESULT hr=S_OK;
	long error=0;
	BSTR berror_msg;

	DLogWMsg(L"Entering AbortThread\n");
	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	berror_msg = SysAllocString(L"");
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPILaunch, (void**) &pLaunch);
	delete arg->ppStream;

	// Wait for the event to be signalled either by normal job termination or Abort(...)
	WaitForSingleObject(arg->hAbortEvent, INFINITE);

	// If the job is being aborted then terminate the launched process
	if (!arg->bNormalExit)
	{
		hr = pLaunch->Abort(&error, &berror_msg);
	}

	pLaunch->Release();
	CoUninitialize();

	// The same event used to signal an abort event is used here to signal when the abort call has completed
	SetEvent(arg->hAbortEvent);
	DLogWMsg(L"Exiting AbortThread\n");
}

// Function name	: PutInputThread
// Description	    : 
// Return type		: void 
// Argument         : PutInputThreadArg *arg
void PutInputThread(PutInputThreadArg *arg)
{
	IMPILaunch *pLaunch=NULL;
	HRESULT hr=S_OK;
	HANDLE hEvent[3];
	long error=0;
	BSTR berror_msg;
	DWORD ret_val;

	DLogWMsg(L"Entering PutInputThread, job %s\n", arg->pJob->m_pszJobID);
	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	berror_msg = SysAllocString(L"");
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPILaunch, (void**) &pLaunch);
	delete arg->ppStream;

	hEvent[0] = arg->pJob->m_hAbortEvent; // This thread should exit on an abort event - normal or abnormal termination
	hEvent[1] = arg->pJob->m_hInputEvent;
	hEvent[2] = arg->pJob->m_hBreakEvent;

	while (true)
	{
		ret_val = WaitForMultipleObjects(3, hEvent, FALSE, INFINITE);
		if (ret_val == WAIT_OBJECT_0)
			// Signalling the abort event or any abnormal wait termination causes a break after which the thread exits
			break;
		if (ret_val == WAIT_OBJECT_0+1)
		{
			if (WaitForSingleObject(arg->pJob->m_hInputMutex, 10000) == WAIT_OBJECT_0)
			{
				VARIANT vInput;
				SAFEARRAYBOUND bound;
				void *pBuf;
				InputNode *node;
				
				VariantInit(&vInput);
				vInput.vt = VT_UI1 | VT_ARRAY;
				bound.lLbound = 0;
				bound.cElements = arg->pJob->m_pInList->size;
				
				// Create an array to put the data in.
				vInput.parray = SafeArrayCreate(VT_UI1, 1, &bound);
				
				// Copy the data in the list to the array
				if (arg->pJob->m_pInList->size > 0)
				{
					SafeArrayAccessData(vInput.parray, &pBuf);
					memcpy(pBuf, arg->pJob->m_pInList->data, arg->pJob->m_pInList->size);
					SafeArrayUnaccessData(vInput.parray);
				}
				
				// Remove the node that has just been copied
				node = arg->pJob->m_pInList;
				
				arg->pJob->m_pInList = arg->pJob->m_pInList->next;
				if (arg->pJob->m_pInList == NULL)
				{
					// Only reset the event if there are no more nodes in the list
					arg->pJob->m_pInListTail = NULL;
					ResetEvent(arg->pJob->m_hInputEvent);
				}
				
				delete node;
				
				ReleaseMutex(arg->pJob->m_hInputMutex);

				// Send the data along to the launched process
				error = 0;
				hr = pLaunch->PutInteractiveInput(vInput, &error, &berror_msg);
				if (FAILED(hr))
				{
					LogMsg(TEXT("PutInputThread:PutInteractiveInput failed: %d\n"), hr);
					break;
				}
				if (error)
				{
					LogWMsg(L"PutInputThread:PutInteractiveInput failed: %d %s\n", error ,berror_msg);
					break;
				}

				// Free the allocated array
				VariantClear(&vInput);
			}
		}
		if (ret_val == WAIT_OBJECT_0+2)
		{
			pLaunch->SendBreak(&error, &berror_msg);
			if (FAILED(hr))
			{
				LogMsg(TEXT("PutInputThread:SendBreak failed: %d\n"), hr);
				break;
			}
			if (error)
			{
				LogWMsg(L"PutInputThread:SendBreak failed: %d %s\n", error ,berror_msg);
				break;
			}
		}
	}

	pLaunch->Release();
	CoUninitialize();
	SysFreeString(berror_msg);
	DLogWMsg(L"Exiting PutInputThread, job %s\n", arg->pJob->m_pszJobID);
}

// Function name	: CreateEnvironmentString
// Description	    : 
// Return type		: BSTR 
// Argument         : BSTR bJobID
// Argument         : long nIproc
// Argument         : long nNproc
// Argument         : BSTR bRootHost
// Argument         : long nRootPort
// Argument         : BSTR bExtra
BSTR CreateEnvironmentString(BSTR bJobID, long nIproc, long nNproc, BSTR bRootHost, long nRootPort, BSTR bExtra, BSTR bNic, long nSMPLow, long nSMPHigh, BSTR bEnv)
{
	WCHAR wBuffer[256];

	if (wcslen(bExtra) > 0)
	{
		swprintf(wBuffer, L"MPICH_JOBID=%s|MPICH_IPROC=%d|MPICH_NPROC=%d|MPICH_ROOTHOST=%s|MPICH_ROOTPORT=%d|MPICH_EXTRA=%s",
			bJobID, nIproc, nNproc, bRootHost, nRootPort, bExtra);
	}
	else
	{
		swprintf(wBuffer, L"MPICH_JOBID=%s|MPICH_IPROC=%d|MPICH_NPROC=%d|MPICH_ROOTHOST=%s|MPICH_ROOTPORT=%d",
			bJobID, nIproc, nNproc, bRootHost, nRootPort);
	}
	if (wcslen(bNic) > 0)
	{
		wcscat(wBuffer, L"|MPICH_COMNIC=");
		wcscat(wBuffer, bNic);
	}
	if (nSMPLow != nSMPHigh)
	{
		WCHAR wBuf2[100];
		swprintf(wBuf2, L"|MPICH_SHM_LOW=%d|MPICH_SHM_HIGH=%d", nSMPLow, nSMPHigh);
		wcscat(wBuffer, wBuf2);
	}
	if (wcslen(bEnv) > 0)
	{
		wcscat(wBuffer, L"|");
		wcscat(wBuffer, bEnv);
	}
	DLogWMsg(L"CreateEnvironmentStrings:\n%s\n", wBuffer);

	return SysAllocString(wBuffer);
}

// Function name	: LaunchJobThread
// Description	    : 
// Return type		: void 
// Argument         : LaunchJobThreadArg *arg
void LaunchJobThread(LaunchJobThreadArg *arg)
{
	HRESULT hr = S_OK;
	IMPILaunch *pLaunch = NULL;
	BSTR berror_msg = NULL;
	long error = 0;
	long more = 1;
	VARIANT *v=NULL;
	SAFEARRAYBOUND bound;
	DWORD num_elements=0;
	void *pBuf=NULL;

	DLogWMsg(L"Entering LaunchJobThread, job %s\n", arg->bJobID);
	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr))
	{
		arg->hr = hr;
		Translate_HRError(arg->hr, arg->wErrorMsg, L"LaunchJobThread:CoInitialize failed ");
		LogWMsg(arg->wErrorMsg);
		CoUninitialize();
		SetEvent(arg->hReadyEvent);
		return;
	}

	//hr = CoCreateInstance(CLSID_MPILaunch, NULL, CLSCTX_SERVER, IID_IMPILaunch, (void**)&pLaunch);
	hr = Connect(pLaunch);

	if (SUCCEEDED(hr))
	{
		berror_msg = SysAllocString(L"");

		BSTR bEnv, bDir;
		if (arg->pJob->m_bNoMPI)
			bEnv = SysAllocString(L"");
		else
			bEnv = CreateEnvironmentString(arg->bJobID, arg->nIproc, arg->nNproc, 
				arg->bRootHost, arg->nRootPort, arg->bExtra, arg->bNic, arg->shmLow, arg->shmHigh, arg->pJob->m_pszEnv);
		bDir = SysAllocString(arg->pJob->m_pszWorkingDir);

		arg->hr = pLaunch->LaunchInteractiveMPIProcess(
			arg->bCmdLine, bDir, bEnv, arg->bAccount, arg->bPassword, 
			&arg->pid, &error, &berror_msg);

		SysFreeString(bEnv);
		SysFreeString(bDir);

		if (FAILED(arg->hr))
		{
			Translate_HRError(arg->hr, arg->wErrorMsg, L"LaunchJobThread:LaunchInteractiveMPIProcess failed ");
			LogWMsg(L"DCOM failure: %s\n", arg->wErrorMsg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			SetEvent(arg->hReadyEvent);
			return;
		}
		if (error)
		{
			swprintf(arg->wErrorMsg, L"LaunchJobThread:%s", berror_msg);
			LogWMsg(arg->wErrorMsg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			CoUninitialize();
			arg->hr = E_FAIL;
			//arg->hr = error;
			SetEvent(arg->hReadyEvent);
			return;
		}

		if (!arg->pJob->m_bNoMPI)
		{
			/*
			if (arg->nIproc == 0)
			{
				// Process zero sets this global event when it has acquired a port and written it to 
				// the temporary file described by bExtra
				WCHAR wBuffer[100];
				wcscpy(wBuffer, arg->bJobID);
				wcscat(wBuffer, L".MPIEvent");
				HANDLE hEvent = CreateEventW(NULL, TRUE, FALSE, wBuffer);
				if (WaitForSingleObject(hEvent, 10000) == WAIT_TIMEOUT)
				{
					arg->hr = E_FAIL;
					LogWMsg(L"Wait for '%s' timed out\n", wBuffer);
					swprintf(arg->wErrorMsg, L"LaunchJobThread:Wait for '%s' timed out", wBuffer);
					SetEvent(arg->hReadyEvent);
					CloseHandle(hEvent);
					CoUninitialize();
					return;
				}
				CloseHandle(hEvent);
			}
			/*/
			if (arg->nIproc == 0)
			{
				// Open the file and read the port number written by the first process
				// The '\n' character signals that the number is complete
				// I could just write the number in binary format and read the first 4 bytes
				HANDLE hFile = CreateFileW(arg->bExtra, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
				if (hFile == INVALID_HANDLE_VALUE)
				{
					arg->hr = E_FAIL;
					Translate_Error(GetLastError(), arg->wErrorMsg, L"LaunchJobThread:CreateFile failed ");
					LogWMsg(arg->wErrorMsg);
					SetEvent(arg->hReadyEvent);
					CoUninitialize();
					return;
				}
				
				DWORD num_read = 0;
				char pBuffer[100];
				pBuffer[0] = '\0';
				char *pChar = pBuffer;
				clock_t cStart = clock();
				while (true)
				{
					num_read = 0;
					if (!ReadFile(hFile, pChar, 100, &num_read, NULL))
					{
						arg->hr = E_FAIL;
						Translate_Error(GetLastError(), arg->wErrorMsg, L"LaunchJobThread:ReadFile failed ");
						LogWMsg(arg->wErrorMsg);
						SetEvent(arg->hReadyEvent);
						CoUninitialize();
						return;
					}
					if (num_read == 0)
					{
						if (clock() - cStart > 10 * CLOCKS_PER_SEC)
						{
							arg->hr = E_FAIL;
							LogWMsg(L"Wait for process 0 to write port to temporary file timed out\n");
							swprintf(arg->wErrorMsg, L"LaunchJobThread:Wait for process 0 to write port to temporary file timed out.");
							SetEvent(arg->hReadyEvent);
							CoUninitialize();
							return;
						}
						Sleep(100);
					}
					else
					{
						for (unsigned int i=0; i<num_read; i++)
						{
							if (*pChar == '\n')
								break;
							pChar ++;
						}
						if (*pChar == '\n')
							break;
					}
				}
				CloseHandle(hFile);
			}
			//*/
		}

		// Every smp process shares this event, so resetting it here is not safe
		ResetEvent(arg->pJob->m_hAbortEvent);

		AbortThreadArg *pcArg = new AbortThreadArg;
		pcArg->hAbortEvent = arg->pJob->m_hAbortEvent;
		pcArg->bNormalExit = false;
		pcArg->ppStream = new IStream*;

		// marshall pLaunch to a thread which waits for an abort event
		arg->hr = CoMarshalInterThreadInterfaceInStream(IID_IMPILaunch, pLaunch, pcArg->ppStream);
		if (FAILED(arg->hr))
		{
			pLaunch->Abort(&error, &berror_msg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			LogWMsg(L"CoMarshalInterThreadInterfaceInStream failed in LaunchJob thread.\n");
			Translate_HRError(arg->hr, arg->wErrorMsg, L"LaunchJobThread:CoMarshalInterThreadInterfaceInStream failed ");
			SetEvent(arg->hReadyEvent);
			CoUninitialize();
			delete pcArg;
			return;
		}

		DWORD dwThreadID;
		HANDLE hAbortThread = CreateThread(
			NULL, 0, 
			(LPTHREAD_START_ROUTINE)AbortThread,
			pcArg, 
			0, &dwThreadID);

		if (hAbortThread == NULL)
		{
			error = GetLastError();
			Translate_Error(error, arg->wErrorMsg, L"LaunchJobThread:CreateThread(AbortThread) failed ");
			LogMsg(TEXT("CreateThread failed: error %d\n"), error);
			pLaunch->Abort(&error, &berror_msg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			SetEvent(arg->hReadyEvent);
			CoUninitialize();
			delete pcArg;
			return;
		}

		PutInputThreadArg *piArg = new PutInputThreadArg;
		piArg->pJob = arg->pJob;
		piArg->ppStream = new IStream*;
		// marshall pLaunch to a thread which waits for user input to be redirected
		arg->hr = CoMarshalInterThreadInterfaceInStream(IID_IMPILaunch, pLaunch, piArg->ppStream);
		if (FAILED(arg->hr))
		{
			pLaunch->Abort(&error, &berror_msg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			Translate_HRError(arg->hr, arg->wErrorMsg, L"LaunchJobThread:CoMarshalInterThreadInterfaceInStream failed ");
			LogWMsg(L"CoMarshalInterThreadInterfaceInStream failed in LaunchJob thread.\n");
			SetEvent(arg->hReadyEvent);
			CoUninitialize();
			delete piArg;
			return;
		}

		HANDLE hPutInputThread = CreateThread(
			NULL, 0, 
			(LPTHREAD_START_ROUTINE)PutInputThread,
			piArg, 
			0, &dwThreadID);

		if (hPutInputThread == NULL)
		{
			error = GetLastError();
			Translate_Error(error, arg->wErrorMsg, L"LaunchJobThread:CreateThread(PutInputThread) failed ");
			LogMsg(TEXT("CreateThread faild: error %d\n"), error);
			pLaunch->Abort(&error, &berror_msg);
			pLaunch->Release();
			SysFreeString(berror_msg);
			SetEvent(arg->hReadyEvent);
			CoUninitialize();
			delete piArg;
			return;
		}

		// Signal that the process has been launched
		SetEvent(arg->hReadyEvent);

		// Collect the output from the launched process
		v = new VARIANT;
		v->vt = VT_UI1 | VT_ARRAY;
		bound.lLbound = 0;
		bound.cElements = 0;
		v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
		
		DLogWMsg(L"LaunchThread: Getting data from launched process.\n");
		while (more == 1)
		{
			error = 0;
			hr = pLaunch->GetInteractiveOutput(v, &more, &error, &berror_msg);
			if (FAILED(hr))
			{
				WCHAR error_msg[256];
				Translate_HRError(hr, error_msg, L"LaunchJobThread:GetInteractiveOutput failed ");
				LogWMsg(error_msg);
				/*
				VariantClear(v);
				delete v;
				pLaunch->Release();
				SysFreeString(berror_msg);
				CoUninitialize();
				return;
				//*/
				break;
			}
			if (error)
			{
				LogWMsg(L"LaunchJobThread: Error %d: %s\n", error, berror_msg);
				/*
				VariantClear(v);
				delete v;
				pLaunch->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				return;
				//*/
				break;
			}
			if (v->parray != NULL)
			{
				num_elements = v->parray->rgsabound[0].cElements;
				if (num_elements > 0)
				{
					DLogWMsg(L"LaunchThread: %d bytes about to be inserted in list\n", num_elements);
					// Add the data to the output queue
					OutputNode *n = new OutputNode;
					n->data = new unsigned char[num_elements];
					SafeArrayAccessData(v->parray, &pBuf);
					memcpy(n->data, pBuf, num_elements);
					SafeArrayUnaccessData(v->parray);
					n->size = num_elements;
					n->from = arg->nIproc;
					n->next = NULL;
					if (WaitForSingleObject(arg->pJob->m_hOutputMutex, 10000) == WAIT_OBJECT_0)
					{
						if (arg->pJob->m_pOutListTail == NULL)
						{
							arg->pJob->m_pOutList = arg->pJob->m_pOutListTail = n;
						}
						else
						{
							arg->pJob->m_pOutListTail->next = n;
							arg->pJob->m_pOutListTail = n;
						}
						SetEvent(arg->pJob->m_hOutputEvent);
						ReleaseMutex(arg->pJob->m_hOutputMutex);
					}
					else
					{
						delete n->data;
						delete n;
					}
					
					// Destroy the array.
					SafeArrayDestroy(v->parray);
					// Create a new and empty array.
					v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				}
				else
				{
					SafeArrayDestroy(v->parray);
					v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				}
			}
			else
			{
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
			}
		}
		DLogWMsg(L"LaunchThread: no more data coming from YLaunchCOM.\n");
		
		VariantClear(v);
		delete v;
		
		pLaunch->Release();
		SysFreeString(berror_msg);

		// Signal and wait for the Abort and PutInput threads to terminate
		// Do this before inserting the termination node in the output list
		// to guarantee that the threads have terminated before this object
		// is released.  Note: Abort will fail after these threads close.
		pcArg->bNormalExit = true;
		SetEvent(pcArg->hAbortEvent);
		if (WaitForSingleObject(hAbortThread, 5000) == WAIT_TIMEOUT)
			LogMsg("LaunchJobThread:AbortThread failed to terminate\n");
		if (WaitForSingleObject(hPutInputThread, 5000) == WAIT_TIMEOUT)
			LogMsg("LaunchJobThread:PutInputThread failed to terminate\n");
		CloseHandle(hAbortThread);
		CloseHandle(hPutInputThread);
		delete pcArg;

		// Insert a termination node in the list
		OutputNode *n = new OutputNode;
		n->from = arg->nIproc;
		n->size = 0; // <------ zero indicates the end of the stream
		n->data = NULL;
		n->next = NULL;
		if (WaitForSingleObject(arg->pJob->m_hOutputMutex, 10000) == WAIT_OBJECT_0)
		{
			if (arg->pJob->m_pOutListTail == NULL)
				arg->pJob->m_pOutList = arg->pJob->m_pOutListTail = n;
			else
			{
				arg->pJob->m_pOutListTail->next = n;
				arg->pJob->m_pOutListTail = n;
			}
			SetEvent(arg->pJob->m_hOutputEvent);
			ReleaseMutex(arg->pJob->m_hOutputMutex);
		}
		else
		{
			delete n;
		}

		CoUninitialize();
	}
	else
	{
		arg->hr = hr;
		Translate_HRError(hr, arg->wErrorMsg, L"LaunchJobThread:CoCreateInstance(IMPILaunch) failed ");
		LogWMsg(arg->wErrorMsg);
		CoUninitialize();
		SetEvent(arg->hReadyEvent);
		return;
	}
	DLogWMsg(L"Exiting LaunchJobThread, job %s\n", arg->bJobID);
}

// Function name	: vSplit
// Description	    : 
// Return type		: VARIANT* 
// Argument         : VARIANT *v
// Argument         : VARIANT *&yours
// Argument         : WCHAR *host
// Argument         : long *iproc
VARIANT* vSplit(VARIANT *v, VARIANT *&yours, WCHAR *host, long *iproc, long *nSMPProcs)
{
	HRESULT hr;
	VARIANT *mine=NULL;
	int size = 0, half = 0, index, i;
	bool odd = true;
	BSTR *bptr;
	long vindex;
	long yindex;
	long mindex;
	SAFEARRAYBOUND bound;
	BSTR *yptr;
	BSTR *mptr;

	yours = NULL;
	host[0] = L'\0';
	*iproc = 0;
	if ((v == NULL) || (v->parray == NULL))
		return NULL;
	
	size = v->parray->rgsabound->cElements; 

	if (size == 0)
		return NULL;

	vindex = 0;
	SafeArrayPtrOfIndex(v->parray, &vindex, (void**)&bptr);
	*iproc = _wtol(*bptr);
	while (iswdigit(**bptr)) (*bptr)++; // advance over the number
	while (iswspace(**bptr)) (*bptr)++; // advance over any white space
	*nSMPProcs = _wtol(*bptr);
	while (iswdigit(**bptr)) (*bptr)++; // advance over the number
	while (iswspace(**bptr)) (*bptr)++; // advance over any white space
	wcscpy(host, *bptr);

	half = size / 2;
	odd = size & 0x1;
	if (size > 1)
	{
		yours = new VARIANT;

		index = half;
		if (odd)
			index++;

		yours->vt = VT_BSTR | VT_ARRAY;
		bound.lLbound = 0;
		bound.cElements = half;
		yours->parray = SafeArrayCreate(VT_BSTR, 1, &bound);
		for (i=0; i<half; i++)
		{
			yindex = i;
			hr = SafeArrayPtrOfIndex(yours->parray, &yindex, (void**)&yptr);
			if (FAILED(hr))
			{
				VariantClear(yours);
				delete yours;
				yours = NULL;
				LogWMsg(L"vSplit: SafeArrayPtrOfIndex() failed.\n");
				return NULL;
			}
			vindex = index;
			hr = SafeArrayPtrOfIndex(v->parray, &vindex, (void**)&bptr);
			if (FAILED(hr))
			{
				VariantClear(yours);
				delete yours;
				yours = NULL;
				LogWMsg(L"vSplit: SafeArrayPtrOfIndex() failed.\n");
				return NULL;
			}
			*yptr = SysAllocString(*bptr);
			index++;
		}
	}

	if (size > 2)
	{
		mine = new VARIANT;

		if (!odd)
			half--;

		mine->vt = VT_BSTR | VT_ARRAY;
		bound.lLbound = 0;
		bound.cElements = half;
		mine->parray = SafeArrayCreate(VT_BSTR, 1, &bound);
		for (i=1; i<half+1; i++)
		{
			mindex = i-1;
			hr = SafeArrayPtrOfIndex(mine->parray, &mindex, (void**)&mptr);
			if (FAILED(hr))
			{
				VariantClear(mine);
				delete mine;
				LogWMsg(L"vSplit: SafeArrayPtrOfIndex() failed.\n");
				return NULL;
			}
			vindex = i;
			hr = SafeArrayPtrOfIndex(v->parray, &vindex, (void**)&bptr);
			if (FAILED(hr))
			{
				VariantClear(mine);
				delete mine;
				LogWMsg(L"vSplit: SafeArrayPtrOfIndex() failed.\n");
				return NULL;
			}
			*mptr = SysAllocString(*bptr);
		}
	}

	return mine;
}

// Function name	: AddNumbersToNames
// Description	    : nIproc and nSMPProcs are prepended to each host name as a way to keep track of who is who
// Return type		: void 
// Argument         : VARIANT &vHosts
void AddNumbersToNames(VARIANT &vHosts, VARIANT &vSMPInfo, long nSecondHostIproc)
{
	if (vHosts.vt == (VT_ARRAY | VT_BSTR))
	{
		BSTR *bptr;
		long *pLong;
		long size, vindex, curIproc = nSecondHostIproc;
		WCHAR wbuffer[256];

		size = vHosts.parray->rgsabound->cElements; 

		if (size == 0)
			return;

		for (vindex=0; vindex<size; vindex++)
		{
			SafeArrayPtrOfIndex(vHosts.parray, &vindex, (void**)&bptr);
			SafeArrayPtrOfIndex(vSMPInfo.parray, &vindex, (void**)&pLong);
			swprintf(wbuffer, L"%d %d %s", curIproc, *pLong, *bptr);
			SysReAllocString(bptr, wbuffer);
			curIproc = curIproc + *pLong;
		}
	}
}

// Function name	: CMPIJob::CreateJob
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT vhosts
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::CreateJob(VARIANT vHosts, VARIANT vSMPInfo, long nSMPProcs, BSTR bComNic, long *error, BSTR *berror_msg)
{
	int i;
	HKEY hKey=NULL, hJobKey=NULL, hProcessesKey=NULL, hTempFilesKey=NULL;
	DWORD ret_val=0, result=0;
	DWORD job_number=0, type, num_bytes = sizeof(DWORD);
	HANDLE hMutex = CreateMutex(NULL, FALSE, TEXT("MPIJobNumberMutex"));
	WCHAR wbuffer[256];
	HRESULT hr=S_OK;

	try{
	// The following two statements are not thread safe.
	if (m_bCreateJobCalled)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"CreateJob:CreateJob can only be called once.");
		return S_OK;
	}
	m_bCreateJobCalled = true;

	// Synchronize access to the job number in the registry
	if ((ret_val = WaitForSingleObject(hMutex, 3000)) != WAIT_OBJECT_0)
	{
		if (ret_val == WAIT_TIMEOUT)
		{
			*error = GetLastError();
			SysReAllocString(berror_msg, L"CreateJob:Timeout waiting for MPIJobNumberMutex");
			LogMsg(TEXT("Timeout waiting for MPIJobNumberMutex\n"));
			CloseHandle(hMutex);
			m_bInconsistentState = true;
			return S_OK;
		}
		else
		{
			*error = GetLastError();
			SysReAllocString(berror_msg, L"CreateJob:Error waiting for MPIJobNumberMutex");
			LogMsg(TEXT("Error waiting for MPIJobNumberMutex\n"));
			CloseHandle(hMutex);
			m_bInconsistentState = true;
			return S_OK;
		}
	}

	// Open the MPICH root key
	if ((ret_val = RegOpenKeyEx(
			HKEY_LOCAL_MACHINE, 
			MPICHKEY,
			0, KEY_READ | KEY_WRITE, &hKey)) != ERROR_SUCCESS)
	{
		ReleaseMutex(hMutex);
		CloseHandle(hMutex);
		*error = ret_val;
		LogMsg(TEXT("Open MPICH registry key failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Open MPICH registry key failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	// Read the job number
	if ((ret_val = RegQueryValueEx(hKey, TEXT("Job Number"), 0, &type, (BYTE *)&job_number, &num_bytes)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		ReleaseMutex(hMutex);
		CloseHandle(hMutex);
		*error = ret_val;
		LogMsg(TEXT("Query Job Number in the registry failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Query Job Number in the registry failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	// Increment the job number and write it back to the registry
	job_number++;
	if ((ret_val = RegSetValueEx(hKey, TEXT("Job Number"), 0, REG_DWORD, (CONST BYTE *)&job_number, sizeof(DWORD))) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		ReleaseMutex(hMutex);
		CloseHandle(hMutex);
		*error = ret_val;
		LogMsg(TEXT("Set Job Number in the registry failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Set Job Number in the registry failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	ReleaseMutex(hMutex);
	CloseHandle(hMutex);

	// Create a job ID string based on the job number and the host name.
	WCHAR pszHost[100];
	DWORD size = 100;
	GetComputerNameW(pszHost, &size);

	swprintf(m_pszJobID, L"%s.%d", pszHost, job_number);

	size = 100;
	hr = CoImpersonateClient();
	if (FAILED(hr))
		LogMsg(TEXT("CreateJob:CoImpersonateClient failed\n"));
	GetUserNameW(m_pszUser, &size);
	hr = CoRevertToSelf();
	if (FAILED(hr))
		LogMsg(TEXT("CreateJob:CoRevertToSelf failed\n"));
	if (wcslen(m_pszUser) < 1)
		swprintf(m_pszUser, L"DefaultUser");

	// Create the registry entries for this job
	swprintf(wbuffer, L"JOBS\\%s\\%s", m_pszUser, m_pszJobID);
	if ((ret_val = RegCreateKeyExW(
		hKey,
		wbuffer,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hJobKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the job registry key failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Creating the job registry key failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	if ((ret_val = RegCreateKeyEx(
		hJobKey,
		TEXT("Processes"),
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hProcessesKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		RegCloseKey(hJobKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the Processes key failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Creating the Processes registry key failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	if ((ret_val = RegCreateKeyEx(
		hJobKey,
		TEXT("TempFiles"),
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hTempFilesKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		RegCloseKey(hJobKey);
		RegCloseKey(hProcessesKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the TempFiles key failed\n"));
		SysReAllocString(berror_msg, L"CreateJob:Creating the TempFiles registry key failed");
		m_bInconsistentState = true;
		return S_OK;
	}

	RegCloseKey(hProcessesKey);
	hProcessesKey = NULL;
	RegCloseKey(hTempFilesKey);
	hTempFilesKey = NULL;
	RegCloseKey(hJobKey);
	hJobKey = NULL;
	RegCloseKey(hKey);
	hKey = NULL;
	}catch(...)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"CreateJob:Exception caught while accessing registry");
		if (hProcessesKey != NULL) RegCloseKey(hProcessesKey);
		if (hTempFilesKey != NULL) RegCloseKey(hTempFilesKey);
		if (hJobKey != NULL) RegCloseKey(hJobKey);
		if (hKey != NULL) RegCloseKey(hKey);
		CloseHandle(hMutex);
		m_bInconsistentState = true;
		return S_OK;
	}

	try{
	// Count up all the processes that will be created
	m_nNproc = nSMPProcs;
	if ((vSMPInfo.vt != VT_EMPTY) && (vSMPInfo.parray != NULL))
	{
		long *pLong;
		for (long size=0; size < (long)vSMPInfo.parray->rgsabound->cElements; size++)
		{
			SafeArrayPtrOfIndex(vSMPInfo.parray, &size, (void**)&pLong);
			m_nNproc = m_nNproc + *pLong;
		}
	}

	// Calculate how many threads to be created
	m_nNumThreads = 0;
	if ((vHosts.vt != VT_EMPTY) && (vHosts.parray != NULL))
	{
		unsigned int temp = vHosts.parray->rgsabound->cElements + 1;
		while (temp)
		{
			m_nNumThreads++;
			temp >>= 1;
		}
		m_nNumThreads--;
	}
	m_nOutputStreamsRemaining += m_nNumThreads;

	// Store the nic this process should use
	wcscpy(m_pszNic, bComNic);

	// Store how many processes should be launched on this host
	m_nSMPProcs = nSMPProcs;

	// Create the ExtendJob threads
	if (m_nNumThreads > 0)
	{
		HANDLE *hConnectedEvents = new HANDLE[m_nNumThreads];
		m_hThreads = new HANDLE[m_nNumThreads];
		m_pThreadArgs = new ExtendJobThreadArg[m_nNumThreads];

		AddNumbersToNames(vHosts, vSMPInfo, nSMPProcs);

		VARIANT *mine = &vHosts;
		for (i=0; i<m_nNumThreads; i++)
		{
			m_hThreads[i] = NULL;
			m_pThreadArgs[i].hCommandEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
			m_pThreadArgs[i].hCommandCompletedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
			m_pThreadArgs[i].pJob = this;
			m_pThreadArgs[i].nCommand = 0;
			m_pThreadArgs[i].pCommandData = NULL;
			m_pThreadArgs[i].nResult = S_OK;
			m_pThreadArgs[i].nNproc = m_nNproc;
			m_pThreadArgs[i].wErrorMsg[0] = L'\0';
			mine = vSplit(
				mine, 
				m_pThreadArgs[i].vHosts, 
				m_pThreadArgs[i].pszHost, 
				&m_pThreadArgs[i].nIproc,
				&m_pThreadArgs[i].nSMPProcs);
			DWORD dwThreadID;
			m_hThreads[i] = CreateThread(
				NULL, 0, 
				(LPTHREAD_START_ROUTINE)ExtendJobThread,
				&m_pThreadArgs[i],
				0, &dwThreadID);
			if (m_hThreads[i] == NULL)
			{
				int terror = GetLastError();
				LogMsg(TEXT("CreateJob:CreateThread[%d] failed: error %d\n"), i, terror);
				Translate_Error(terror, m_pThreadArgs[i].wErrorMsg, L"CreateJob:CreateThread failed ");
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
			hConnectedEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
		}

		// Wait for all the threads to signal they have connected
		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hConnectedEvents, TRUE, INFINITE);
		delete hConnectedEvents;

		*error = 0;
		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				WCHAR wBuffer[1024];
				*error = m_pThreadArgs[i].nResult;
				if (wcslen(m_pThreadArgs[i].wErrorMsg) < 1)
					Translate_HRError(*error, m_pThreadArgs[i].wErrorMsg);
				swprintf(wBuffer, L"CreateJob:%s", m_pThreadArgs[i].wErrorMsg);
				SysReAllocString(berror_msg, wBuffer);
				LogWMsg(wBuffer);
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
			}
		}
		if (*error)
		{
			m_bInconsistentState = true;
			return S_OK;
		}
	}
	}catch(...)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"CreateJob:Exception caught while connecting to remote hosts");
		m_bInconsistentState = true;
		return S_OK;
	}
	return S_OK;
}

// Function name	: CMPIJob::ExtendJob
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bJobID
// Argument         : VARIANT vHosts
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::ExtendJob(long iproc, long nproc, BSTR bJobID, VARIANT vHosts, long nSMPProcs, BSTR bComNic, long *error, BSTR *berror_msg)
{
	int i;
	HKEY hKey=NULL, hJobKey=NULL, hProcessesKey=NULL, hTempFilesKey=NULL;
	DWORD ret_val, result;
	WCHAR wbuffer[256];
	DWORD size = 100;

	try{
	wcscpy(m_pszJobID, bJobID);
	wcscpy(m_pszNic, bComNic);
	m_nIproc = iproc;
	m_nNproc = nproc;
	m_nSMPProcs = nSMPProcs;

	if (m_nNumThreads > 0)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"ExtendJob: class not initialized properly");
		LogMsg(TEXT("ExtendJob: class not initialized properly\n"));
		return S_OK;
	}

	// Open the MPICH root key
	if ((ret_val = RegOpenKeyEx(
			HKEY_LOCAL_MACHINE, 
			MPICHKEY,
			0, KEY_READ | KEY_WRITE, &hKey)) != ERROR_SUCCESS)
	{
		*error = ret_val;
		LogMsg(TEXT("Opening the MPICH registry key failed\n"));
		SysReAllocString(berror_msg, L"ExtendJob:Opening the MPICH registry key failed");
		return S_OK;
	}

	size = 100;
	CoImpersonateClient();
	GetUserNameW(m_pszUser, &size);
	CoRevertToSelf();
	if (wcslen(m_pszUser) < 1)
		swprintf(m_pszUser, L"DefaultUser");

	// Create the registry entries for this job
	swprintf(wbuffer, L"JOBS\\%s\\%s", m_pszUser, m_pszJobID);
	if ((ret_val = RegCreateKeyExW(
		hKey,
		wbuffer,
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hJobKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the job registry key failed\n"));
		SysReAllocString(berror_msg, L"ExtendJob:Creating the job registry key failed");
		return S_OK;
	}

	if ((ret_val = RegCreateKeyEx(
		hJobKey,
		TEXT("Processes"),
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hProcessesKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		RegCloseKey(hJobKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the Processes registry key failed\n"));
		SysReAllocString(berror_msg, L"ExtendJob:Creating the Process registry key failed");
		return S_OK;
	}

	if ((ret_val = RegCreateKeyEx(
		hJobKey,
		TEXT("TempFiles"),
		0, NULL, REG_OPTION_NON_VOLATILE, KEY_READ | KEY_WRITE, NULL,
		&hTempFilesKey,
		&result)) != ERROR_SUCCESS)
	{
		RegCloseKey(hKey);
		RegCloseKey(hJobKey);
		RegCloseKey(hProcessesKey);
		*error = ret_val;
		LogMsg(TEXT("Creating the TempFiles registry key failed\n"));
		SysReAllocString(berror_msg, L"ExtendJob:Creating the TempFiles registry key failed");
		return S_OK;
	}

	RegCloseKey(hKey);
	RegCloseKey(hJobKey);
	RegCloseKey(hProcessesKey);
	RegCloseKey(hTempFilesKey);

	// Calculate how many threads to be created
	m_nNumThreads = 0;
	if ((vHosts.vt != VT_EMPTY) && (vHosts.parray != NULL))
	{
		unsigned int temp = vHosts.parray->rgsabound->cElements + 1;
		while (temp)
		{
			m_nNumThreads++;
			temp >>= 1;
		}
		m_nNumThreads--;
	}
	m_nOutputStreamsRemaining += m_nNumThreads;

	// Create the ExtendJob threads
	if (m_nNumThreads > 0)
	{
		HANDLE *hConnectedEvents = new HANDLE[m_nNumThreads];
		m_hThreads = new HANDLE[m_nNumThreads];
		m_pThreadArgs = new ExtendJobThreadArg[m_nNumThreads];

		VARIANT *mine = &vHosts;
		for (i=0; i<m_nNumThreads; i++)
		{
			m_hThreads[i] = NULL;
			m_pThreadArgs[i].hCommandEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
			m_pThreadArgs[i].hCommandCompletedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
			m_pThreadArgs[i].pJob = this;
			m_pThreadArgs[i].nCommand = 0;
			m_pThreadArgs[i].pCommandData = NULL;
			m_pThreadArgs[i].nResult = S_OK;
			m_pThreadArgs[i].nNproc = m_nNproc;
			mine = vSplit(
				mine, 
				m_pThreadArgs[i].vHosts, 
				m_pThreadArgs[i].pszHost, 
				&m_pThreadArgs[i].nIproc,
				&m_pThreadArgs[i].nSMPProcs);
			DWORD dwThreadID;
			m_hThreads[i] = CreateThread(
				NULL, 0, 
				(LPTHREAD_START_ROUTINE)ExtendJobThread,
				&m_pThreadArgs[i],
				0, &dwThreadID);
			if (m_hThreads[i] == NULL)
			{
				int terror = GetLastError();
				LogMsg(TEXT("CreateThread[%d] failed: error %d\n"), i, terror);
				Translate_Error(terror, m_pThreadArgs[i].wErrorMsg, L"ExtendJob:CreateThread failed ");
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
			hConnectedEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
		}

		// Wait for all the threads to signal they have connected
		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hConnectedEvents, TRUE, INFINITE);
		delete hConnectedEvents;

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult)
			{
				WCHAR wBuffer[1024];
				*error = m_pThreadArgs[i].nResult;
				if (wcslen(m_pThreadArgs[i].wErrorMsg) < 1)
					Translate_HRError(*error, m_pThreadArgs[i].wErrorMsg);
				swprintf(wBuffer, L"ExtendJob[%s]:%s", m_pszLocalHost, m_pThreadArgs[i].wErrorMsg);
				SysReAllocString(berror_msg, wBuffer);
				LogWMsg(wBuffer);
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
			}
		}
		if (*error)
			return S_OK;
	}
	}catch(...)
	{
		*error = 1;
		if (hProcessesKey != NULL) RegCloseKey(hProcessesKey);
		if (hTempFilesKey != NULL) RegCloseKey(hTempFilesKey);
		if (hJobKey != NULL) RegCloseKey(hJobKey);
		if (hKey != NULL) RegCloseKey(hKey);
		SysReAllocString(berror_msg, L"ExtendJob:Exception caught while connecting to remote hosts");
		return S_OK;
	}
	return S_OK;
}

// Function name	: CMPIJob::DestroyJob
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::DestroyJob(long *error, BSTR *berror_msg)
{
	int i;
	HKEY hKey, hKey2;
	WCHAR wBuffer[100];
	HANDLE hMutex;
	LONG ret_val;

	if (m_nNumThreads > 0 && m_pThreadArgs && m_hThreads)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		// Signal each of the threads to destroy the job
		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = NULL;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_EXIT;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
			}
		}

		// Wait for all the threads to terminate
		/*
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);
		delete hEvents;

		// Close all open handles and free allocated memory
		for (i=0; i<m_nNumThreads; i++)
		{
			CloseHandle(m_pThreadArgs[i].hCommandCompletedEvent);
			CloseHandle(m_pThreadArgs[i].hCommandEvent);
			if (m_hThreads[i] != NULL)
				CloseHandle(m_hThreads[i]);
		}
		delete m_pThreadArgs;
		delete m_hThreads;
		m_pThreadArgs = NULL;
		m_hThreads = NULL;
		m_nNumThreads = 0;
		/*/
		DWORD timeout = 2000;
		if (m_nNproc > 1)
			timeout = timeout * (m_nNproc-1);
		if (WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, timeout) != WAIT_OBJECT_0)
		{
			delete hEvents;
			
			// Kill the threads, close all open handles and free allocated memory
			for (i=0; i<m_nNumThreads; i++)
			{
				CloseHandle(m_pThreadArgs[i].hCommandCompletedEvent);
				CloseHandle(m_pThreadArgs[i].hCommandEvent);
				if (m_hThreads[i] != NULL)
				{
					TerminateThread(m_hThreads[i], 0);
					CloseHandle(m_hThreads[i]);
				}
			}
		}
		else
		{
			delete hEvents;
			
			// Close all open handles and free allocated memory
			for (i=0; i<m_nNumThreads; i++)
			{
				CloseHandle(m_pThreadArgs[i].hCommandCompletedEvent);
				CloseHandle(m_pThreadArgs[i].hCommandEvent);
				if (m_hThreads[i] != NULL)
					CloseHandle(m_hThreads[i]);
			}
		}
		delete m_pThreadArgs;
		delete m_hThreads;
		m_pThreadArgs = NULL;
		m_hThreads = NULL;
		m_nNumThreads = 0;
		//*/
	}

	// Syncronize access to the registry
	hMutex = CreateMutex(NULL, FALSE, TEXT("MPIRegistryMutex"));
	if (WaitForSingleObject(hMutex, 10000) == WAIT_OBJECT_0)
	{
		// Open the MPICH root key
		if ((ret_val = RegOpenKeyEx(
			HKEY_LOCAL_MACHINE, 
			MPICHKEY,
			0, KEY_READ | KEY_WRITE, &hKey)) != ERROR_SUCCESS)
		{
			ReleaseMutex(hMutex);
			CloseHandle(hMutex);
			*error = ret_val;
			LogMsg(TEXT("Open MPICH registry key failed\n"));
			SysReAllocString(berror_msg, L"DestroyJob:Open MPICH registry key failed");
			return S_OK;
		}
		
		// Delete the temporary files created by this job
		swprintf(wBuffer, L"JOBS\\%s\\%s\\TempFiles", m_pszUser, m_pszJobID);
		if ((ret_val = RegOpenKeyExW(hKey, wBuffer, 0, KEY_READ | KEY_WRITE, &hKey2)) == ERROR_SUCCESS)
		{
			DWORD num_values = 0, namelen = 256;
			TCHAR name[256];
			// Query how many values there are in the key
			if (RegQueryInfoKey(hKey2, NULL, NULL, NULL, NULL, NULL, NULL, &num_values, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
			{
				if (num_values > 0)
				{
					for (int i=(int)num_values-1; i>=0; i--) // Count backwards so values can be deleted
					{
						namelen = 256;
						if (RegEnumValue(hKey2, i, name, &namelen, NULL, NULL, NULL, NULL) == ERROR_SUCCESS)
						{
							// Try to delete the file
							if (DeleteFile(name))
							{
								RegDeleteValue(hKey2, name);
							}
							else
							{
								int error = GetLastError();
								// If the temporary file does not exist then remove it from the key else leave it there.
								if (error == ERROR_FILE_NOT_FOUND || error == ERROR_PATH_NOT_FOUND)
								{
									RegDeleteValue(hKey2, name);
								}
								else
								{
									TCHAR error_msg[256];
									Translate_Error(error, error_msg);
									LogMsg(TEXT("Unable to delete file(%s) retrieved from the TempFiles Registry key: %s\n"), name, error_msg);
								}
							}
						}
						else
						{
							LogWMsg(L"Unable to enumerate a value entry in the %s\\%s\\TempFiles Registry key.\n", m_pszUser, m_pszJobID);
						}
					}
				}
			}
			else
			{
				LogWMsg(L"Unable to query the number of entries under the %s\\%s\\TempFiles Registry key.\n", m_pszUser, m_pszJobID);
			}
			RegCloseKey(hKey2);
		}
		
		// Delete the TempFiles key
		RegDeleteKeyW(hKey, wBuffer);
		
		// Delete the Processes key
		swprintf(wBuffer, L"JOBS\\%s\\%s\\Processes", m_pszUser, m_pszJobID);
		RegDeleteKeyW(hKey, wBuffer);
		
		// Delete the JobID key
		swprintf(wBuffer, L"JOBS\\%s\\%s", m_pszUser, m_pszJobID);
		RegDeleteKeyW(hKey, wBuffer);
		
		// Delete the User key - This will only succeed when all the job subkeys have been removed
		swprintf(wBuffer, L"JOBS\\%s", m_pszUser);
		RegDeleteKeyW(hKey, wBuffer);
		
		// Delete the JOBS key - This will only succeed when all the User subkeys have been removed
		RegDeleteKey(hKey, TEXT("JOBS"));
		
		ReleaseMutex(hMutex);
		CloseHandle(hMutex);
	}

	return S_OK;
}

// Function name	: CMPIJob::SetExecutable
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bExecutable
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::SetExecutable(BSTR bExecutable, long *error, BSTR *berror_msg)
{
	int i;

	*error = 0;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"SetExecutable:SetExecutable cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	wcscpy(m_pszExe, bExecutable);

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = bExecutable;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_SETEXE;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					WCHAR wBuffer[1024];
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wBuffer, L"SetExecutable:%s", m_pThreadArgs[i].wErrorMsg);
					else
					{
						Translate_Error(m_pThreadArgs[i].nResult, L"SetExecutable failed ");
						wcscpy(wBuffer, m_pThreadArgs[i].wErrorMsg);
					}
					SysReAllocString(berror_msg, wBuffer);
					LogWMsg(wBuffer);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::StartExeTransfer
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::StartExeTransfer(long *error, BSTR *berror_msg)
{
	WCHAR wbuffer[MAX_PATH];
	WCHAR wError[1024];
	HKEY hKey;
	DWORD ret_val;
	int i;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"StartExeTransfer:StartExeTransfer cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	swprintf(wbuffer, LMPICHKEY L"\\JOBS\\%s\\%s\\TempFiles", m_pszUser, m_pszJobID);
	if ((ret_val = RegOpenKeyExW(
		HKEY_LOCAL_MACHINE, 
		wbuffer,
		0,
		KEY_READ | KEY_WRITE,
		&hKey)) != ERROR_SUCCESS)
	{
		*error = ret_val;
		Translate_HRError(ret_val, wError, L"StartExeTransfer:RegOpenKeyExW(MPICH\\JOBS\\...\\TempFiles) failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		return S_OK;
	}

	if (GetTempPathW(MAX_PATH, wbuffer) == 0)
	{
		*error = GetLastError();
		Translate_Error(*error, wError, L"StartExeTransfer:GetTempPath failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		RegCloseKey(hKey);
		return S_OK;
	}
	if (GetTempFileNameW(wbuffer, L"mpi", 0, m_pszExe) == 0)
	{
		*error = GetLastError();
		Translate_Error(*error, wError, L"StartExeTransfer:GetTempFileNameW failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		RegCloseKey(hKey);
		return S_OK;
	}
	m_hTempFile = CreateFileW(
		m_pszExe,
		GENERIC_READ | GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_SEQUENTIAL_SCAN,
		NULL);
	if (m_hTempFile == INVALID_HANDLE_VALUE)
	{
		*error = GetLastError();
		Translate_Error(*error, wError, L"StartExeTransfer:CreateFileW failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		RegCloseKey(hKey);
		return S_OK;
	}

	DWORD dval = 1;
	if ((ret_val = RegSetValueExW(
		hKey, 
		m_pszExe,
		0,
		REG_DWORD,
		(CONST BYTE*)&dval,
		sizeof(DWORD))) != ERROR_SUCCESS)
	{
		*error = ret_val;
		Translate_HRError(ret_val, wError, L"StartExeTransfer:RegSetValueExW(temp file name) failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		RegCloseKey(hKey);
		return S_OK;
	}

	RegCloseKey(hKey);

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = NULL;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_STARTTRANSFER;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"StartExeTransfer:%s", m_pThreadArgs[i].wErrorMsg);
					else
						Translate_Error(m_pThreadArgs[i].nResult, wError, L"StartExeTransfer:ExtendJobThread failed ");
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::TransferExeData
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT vData
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::TransferExeData(VARIANT vData, long *error, BSTR *berror_msg)
{
	int i;
	void *p_buf = NULL;
	HRESULT hr = S_OK;
	WCHAR wError[1024];

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"TransferExeData:TransferExeData cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	if (m_hTempFile == NULL)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"TransferExeData:File not open, probably due to not calling StartExeTransfer first");
		LogMsg(TEXT("TransferExeData:File not open, probably due to not calling StartExeTransfer first\n"));
		return S_OK;
	}

	// Check the input parameter
	if (vData.vt != (VT_UI1 | VT_ARRAY))
	{
		*error = 1;
		SysReAllocString(berror_msg, L"TransferExeData:vData must be SAFEARRAY  (VT_UI1 | VT_ARRAY)");
		LogMsg(TEXT("TransferExeData:vData must be SAFEARRAY  (VT_UI1 | VT_ARRAY)\n"));
		return S_OK;
	}

	// Get a pointer to the input data
	hr = SafeArrayAccessData(vData.parray, &p_buf);

	if (FAILED(hr))
	{
		Translate_HRError(hr, wError, L"TransferExeData:SafeArrayAccessData failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		*error = hr;
		return S_OK;
	}

	// write to file
	DWORD num_written;
	if (!WriteFile(m_hTempFile, p_buf, vData.parray->rgsabound->cElements, &num_written, NULL))
	{
		*error = GetLastError();
		Translate_Error(*error, wError, L"TransferExeData:WriteFile failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		SafeArrayUnaccessData(vData.parray);
		return S_OK;
	}

	// release pointer
	SafeArrayUnaccessData(vData.parray);

	// Signal each of the threads to transfer the data
	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];
		//VARIANT *vTemp = new VARIANT[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				//VariantInit(&vTemp[i]);
				//VariantCopy(&vTemp[i], &vData);
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = &vData;
				//m_pThreadArgs[i].pCommandData = &vTemp[i];
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_TRANSFER;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
				Sleep(0); // Try to ensure that the threads post their sends in order by relinquishing my time slice at each iteration
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			//VariantClear(&vTemp[i]);
			//if (m_pThreadArgs[i].nResult != 0)
			if (FAILED(m_pThreadArgs[i].nResult))
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"TransferExeData:%s", m_pThreadArgs[i].wErrorMsg);
					else
						Translate_Error(*error, wError, L"TransferExeData:ExtendJobThread failed ");
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		//delete vTemp;
		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::FinishExeTransfer
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::FinishExeTransfer(long *error, BSTR *berror_msg)
{
	int i;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"FinishExeTransfer:FinishExeTransfer cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	if (m_hTempFile != NULL)
	{
		CloseHandle(m_hTempFile);
		m_hTempFile = NULL;
	}

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = NULL;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_FINISHTRANSFER;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			//if (m_pThreadArgs[i].nResult != 0)
			if (FAILED(m_pThreadArgs[i].nResult))
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					*error = m_pThreadArgs[i].nResult;
					WCHAR wError[1024];
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"FinishExeTransfer:%s", m_pThreadArgs[i].wErrorMsg);
					else
						Translate_Error(*error, wError, L"FinishExeTransfer:ExtendJobThread failed ");
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

int GetFixedPortFromEnv(LPWSTR pszEnv)
{
	WCHAR wName[MAX_PATH]=L"", wValue[MAX_PATH]=L"";
	WCHAR *pChar;

	pChar = wName;
	while (*pszEnv != L'\0')
	{
		if (*pszEnv == L'=')
		{
			*pChar = L'\0';
			pChar = wValue;
		}
		else
		if (*pszEnv == L'|')
		{
			*pChar = L'\0';
			pChar = wName;
			if (wcscmp(wName, L"MPICH_FIXEDPORT") == 0)
			{
				return _wtoi(wValue);
			}
		}
		else
		{
			*pChar = *pszEnv;
			pChar++;
		}
		pszEnv++;
	}
	*pChar = L'\0';
	if (wcscmp(wName, L"MPICH_FIXEDPORT") == 0)
	{
		return _wtoi(wValue);
	}
	return 0;
}

// Function name	: CMPIJob::LaunchJob
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bCommonArgs
// Argument         : BSTR bAccount
// Argument         : BSTR bPassword
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::LaunchJob(BSTR bCommonArgs, BSTR bAccount, BSTR bPassword, long flags, long *error, BSTR *berror_msg)
{
	int i;
	BSTR pBstrs[5];
	WCHAR wBuffer[MAX_PATH], wTempFileName[MAX_PATH];
	WCHAR wError[1024];
	DWORD size = 100;

	*error = 0;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchJob:LaunchJob cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	m_nLaunchFlags = flags;
	m_bNoMPI = flags & MPI_LAUNCH_OPT_NOMPI;
	m_bFixedPort = flags & MPI_LAUNCH_OPT_FIXEDPORT;

	if (wcslen(m_pszExe) < 1)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchJob:No executable set to be run");
		return S_OK;
	}

	if (!m_bNoMPI && !m_bFixedPort)
	{
		// Create a temporary file used by the first process to write the root port number to
		if (GetTempPathW(MAX_PATH, wBuffer) == 0)
		{
			*error = GetLastError();
			Translate_Error(*error, wError, L"LaunchJob:GetTempPathW failed ");
			SysReAllocString(berror_msg, wError);
			LogWMsg(wError);
			return S_OK;
		}
		if (GetTempFileNameW(wBuffer, L"mpi", 0, wTempFileName) == 0)
		{
			*error = GetLastError();
			Translate_Error(*error, wError, L"LaunchJob:GetTempFileNameW failed ");
			SysReAllocString(berror_msg, wError);
			LogWMsg(wError);
			return S_OK;
		}
	}

	pBstrs[0] = bCommonArgs;
	pBstrs[1] = bAccount;
	pBstrs[2] = bPassword;

	pBstrs[3] = SysAllocString(m_pszNic);

	// Start the thread to launch the root process
	// Read in the root port number
	LaunchJobThreadArg *pArg = new LaunchJobThreadArg;
	pArg->wErrorMsg[0] = L'\0';
	pArg->hr = S_OK;
	pArg->bAccount = bAccount;
	// The root process will write the root port number to the file described by wTempFileName
	if (m_bNoMPI || m_bFixedPort)
		pArg->bExtra = SysAllocString(L"");
	else
		pArg->bExtra = SysAllocString(wTempFileName);
	// Why am I allocating these instead of just using them as they are?
	pArg->bJobID = SysAllocString(m_pszJobID);
	pArg->bPassword = bPassword;
	pArg->bRootHost = SysAllocString(m_pszNic);
	pArg->hReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	pArg->nIproc = 0;
	pArg->nNproc = m_nNproc;
	pArg->nRootPort = -1;
	pArg->pid = -1;
	pArg->pJob = this;
	pArg->bNic = m_pszNic;
	pArg->shmLow = 0;
	pArg->shmHigh = m_nSMPProcs - 1;

	// Create the command line
	WCHAR wCmdLine[MAX_PATH];
	if (wcslen(bCommonArgs) > 0)
		swprintf(wCmdLine, L"%s %s", m_pszExe, bCommonArgs);
	else
		wcscpy(wCmdLine, m_pszExe);
	pArg->bCmdLine = SysAllocString(wCmdLine);

	HANDLE hThread;
	DWORD dwThreadID;
	hThread = CreateThread(
		NULL, 0,
		(LPTHREAD_START_ROUTINE)LaunchJobThread,
		pArg,
		0, &dwThreadID);
	if (hThread == NULL)
	{
		*error = GetLastError();
		Translate_Error(*error, wError, L"LaunchJob:CreateThread(LaunchJobThread) failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		return S_OK;
	}

	if (WaitForSingleObject(pArg->hReadyEvent, 7000) == WAIT_TIMEOUT)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchJob:Timed out waiting for process to launch and reply.");

		// Signal the abort thread to kill the process
		PulseEvent(m_hAbortEvent);
		WaitForSingleObject(m_hAbortEvent, 2000);

		return S_OK;
	}

	if (pArg->hr != S_OK)
	{
		*error = 1;
		if (wcslen(pArg->wErrorMsg) > 0)
			swprintf(wError, L"LaunchJob:%s", pArg->wErrorMsg);
		else
			Translate_HRError(pArg->hr, wError, L"LaunchJob:LaunchJobThread failed ");
		SysReAllocString(berror_msg, wError);
		LogWMsg(wError);
		return S_OK;
	}
	/*
	// Add the process to the registry
	HKEY hkey;
	if (RegOpenKey(HKEY_LOCAL_MACHINE, YPROCKEY, &hkey) == ERROR_SUCCESS)
	{
		TCHAR pid_str[100];
		_stprintf(pid_str, TEXT("%d"), psInfo.dwProcessId);
		RegSetValueEx(hkey, pid_str, 0, REG_SZ, (CONST BYTE*)"*", 2*sizeof(TCHAR));
		RegCloseKey(hkey);
	}
	//*/
	m_bLaunchThreadStarted = true;

	long port=-1;
	if (m_bNoMPI)
	{
		pBstrs[4] = SysAllocString(L"");
	}
	else
	{
		if (m_bFixedPort)
		{
			port = GetFixedPortFromEnv(m_pszEnv);
			if (port == 0)
			{
				SysReAllocString(berror_msg, L"Unable to launch a job with a fixed port because the port is not specified in the environment.");
				LogWMsg(*berror_msg);
				return S_OK;
			}
		}
		else
		{
			HANDLE hFile = CreateFileW(wTempFileName, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
			char buffer[11];
			DWORD num_read = 0;
			ReadFile(hFile, buffer, 10, &num_read, NULL);
			CloseHandle(hFile);
			DeleteFileW(wTempFileName);
			buffer[num_read] = '\0';
			port = atol(buffer);
		}
		swprintf(wBuffer, L"%d", port);
		pBstrs[4] = SysAllocString(wBuffer);
	}

	// Launch the rest of the smp processes
	for (i=1; i<m_nSMPProcs; i++)
	{
		LaunchJobThreadArg *pArg = new LaunchJobThreadArg;
		pArg->wErrorMsg[0] = L'\0';
		pArg->hr = S_OK;
		pArg->bAccount = bAccount;
		pArg->bExtra = SysAllocString(L"");
		// Why am I allocating these instead of just using them as they are?
		pArg->bJobID = SysAllocString(m_pszJobID);
		pArg->bPassword = bPassword;
		pArg->bRootHost = SysAllocString(m_pszNic);
		pArg->hReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		pArg->nIproc = i;
		pArg->nNproc = m_nNproc;
		pArg->nRootPort = port;
		pArg->pid = -1;
		pArg->pJob = this;
		pArg->bNic = m_pszNic;
		pArg->shmLow = 0;
		pArg->shmHigh = m_nSMPProcs - 1;
		
		// Create the command line
		WCHAR wCmdLine[MAX_PATH];
		if (wcslen(bCommonArgs) > 0)
			swprintf(wCmdLine, L"%s %s", m_pszExe, bCommonArgs);
		else
			wcscpy(wCmdLine, m_pszExe);
		pArg->bCmdLine = SysAllocString(wCmdLine);
		
		HANDLE hThread;
		DWORD dwThreadID;
		hThread = CreateThread(
			NULL, 0,
			(LPTHREAD_START_ROUTINE)LaunchJobThread,
			pArg,
			0, &dwThreadID);
		if (hThread == NULL)
		{
			*error = GetLastError();
			Translate_Error(*error, wError, L"LaunchJob:CreateThread(LaunchJobThread) failed ");
			SysReAllocString(berror_msg, wError);
			LogWMsg(wError);
			return S_OK;
		}
		
		if (WaitForSingleObject(pArg->hReadyEvent, 7000) == WAIT_TIMEOUT)
		{
			*error = 1;
			SysReAllocString(berror_msg, L"LaunchJob:Timed out waiting for process to launch.");
			
			// Signal the abort thread to kill the processes
			PulseEvent(m_hAbortEvent);
			WaitForSingleObject(m_hAbortEvent, 2000);
			
			return S_OK;
		}
		
		if (pArg->hr != S_OK)
		{
			*error = 1;
			if (wcslen(pArg->wErrorMsg) > 0)
				swprintf(wError, L"LaunchJob:%s", pArg->wErrorMsg);
			else
				Translate_HRError(pArg->hr, wError, L"LaunchJob:LaunchJobThread failed ");
			SysReAllocString(berror_msg, wError);
			LogWMsg(wError);
			return S_OK;
		}
		/*
		// Add the process to the registry
		HKEY hkey;
		if (RegOpenKey(HKEY_LOCAL_MACHINE, YPROCKEY, &hkey) == ERROR_SUCCESS)
		{
		TCHAR pid_str[100];
		_stprintf(pid_str, TEXT("%d"), psInfo.dwProcessId);
		RegSetValueEx(hkey, pid_str, 0, REG_SZ, (CONST BYTE*)"*", 2*sizeof(TCHAR));
		RegCloseKey(hkey);
		}
		//*/
	}

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = pBstrs;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_LAUNCHJOB;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			//if (m_pThreadArgs[i].nResult != 0)
			if (FAILED(m_pThreadArgs[i].nResult))
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					*error = 1;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"LaunchJob:%s", m_pThreadArgs[i].wErrorMsg);
					else
						Translate_Error(m_pThreadArgs[i].nResult, wError, L"LaunchJob:LaunchJobThread failed ");
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::ExtendLaunchJob
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bCommonArgs
// Argument         : BSTR bAccount
// Argument         : BSTR bPassword
// Argument         : BSTR bRootHost
// Argument         : BSTR bRootPort
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::ExtendLaunchJob(BSTR bCommonArgs, BSTR bAccount, BSTR bPassword, BSTR bRootHost, BSTR bRootPort, long flags, long *error, BSTR *berror_msg)
{
	int i;
	BSTR pBstrs[5];
	DWORD size = 100;
	WCHAR wError[1024];

	*error = 0;

	if (wcslen(m_pszExe) < 1)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"ExtendLaunchJob:No executable set to be run");
		return S_OK;
	}

	m_bNoMPI = flags & MPI_LAUNCH_OPT_NOMPI;
	m_bFixedPort = flags & MPI_LAUNCH_OPT_FIXEDPORT;
	m_nLaunchFlags = flags;

	for (i=0; i<m_nSMPProcs; i++)
	{
		// Start a thread to launch the process associated with this node of the job
		LaunchJobThreadArg *pArg = new LaunchJobThreadArg;
		pArg->wErrorMsg[0] = L'\0';
		pArg->hr = S_OK;
		pArg->bAccount = bAccount;
		pArg->bExtra = SysAllocString(L"");
		pArg->bJobID = SysAllocString(m_pszJobID);
		pArg->bPassword = bPassword;
		pArg->bRootHost = bRootHost;
		pArg->hReadyEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		pArg->nIproc = m_nIproc + i;
		pArg->nNproc = m_nNproc;
		pArg->nRootPort = _wtol(bRootPort);
		pArg->pid = -1;
		pArg->pJob = this;
		pArg->bNic = m_pszNic;
		pArg->shmLow = m_nIproc;
		pArg->shmHigh = m_nIproc + m_nSMPProcs - 1;
		
		// Create the command line
		WCHAR wCmdLine[MAX_PATH];
		//swprintf(wCmdLine, L"cmd.exe /c %s %s", m_pszExe, bCommonArgs);
		swprintf(wCmdLine, L"%s %s", m_pszExe, bCommonArgs);
		pArg->bCmdLine = SysAllocString(wCmdLine);
		
		HANDLE hThread;
		DWORD dwThreadID;
		hThread = CreateThread(
			NULL, 0,
			(LPTHREAD_START_ROUTINE)LaunchJobThread,
			pArg,
			0, &dwThreadID);
		if (hThread == NULL)
		{
			*error = GetLastError();
			Translate_Error(*error, wError, L"ExtendLaunchJob:CreateThread(LaunchJobThread) failed ");
			SysReAllocString(berror_msg, wError);
			LogWMsg(wError);
			return S_OK;
		}
		
		if (WaitForSingleObject(pArg->hReadyEvent, 7000) == WAIT_TIMEOUT)
		{
			*error = 1;
			SysReAllocString(berror_msg, L"ExtendLaunchJob:Timed out waiting for process to launch.");
			return S_OK;
		}
		if (pArg->hr != S_OK)
		{
			*error = pArg->hr;
			if (wcslen(pArg->wErrorMsg) > 0)
				swprintf(wError, L"ExtendLaunchJob[%s]:%s", m_pszLocalHost, pArg->wErrorMsg);
			else
			{
				WCHAR wTemp[200];
				swprintf(wTemp, L"ExtendLaunchJob[%s]:LaunchJobThread failed ", m_pszLocalHost);
				Translate_HRError(*error, wError, wTemp);
			}
			SysReAllocString(berror_msg, wError);
			return S_OK;
		}
	}

	// Signal the threads to launch processes on their nodes also
	pBstrs[0] = bCommonArgs;
	pBstrs[1] = bAccount;
	pBstrs[2] = bPassword;
	pBstrs[3] = bRootHost;
	pBstrs[4] = bRootPort;

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = pBstrs;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_LAUNCHJOB;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			//if (m_pThreadArgs[i].nResult != 0)
			if (FAILED(m_pThreadArgs[i].nResult))
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"ExtendLaunchJob[%s]:%s", m_pszLocalHost, m_pThreadArgs[i].wErrorMsg);
					else
					{
						WCHAR wTemp[200];
						swprintf(wTemp, L"ExtendLaunchJob[%s]:LaunchJobThread failed ", m_pszLocalHost);
						Translate_Error(*error, wError, wTemp);
					}
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::GetJobOutput
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT *vOutput
// Argument         : long *from
// Argument         : long *more
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::GetJobOutput(VARIANT *vOutput, long *from, long *more, long *error, BSTR *berror_msg)
{
	SAFEARRAYBOUND bound;
	void *pBuf;
	OutputNode *node;

	try{
	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"GetJobOutput:GetJobOutput cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	if (vOutput->parray != NULL)
	{
		SafeArrayDestroy(vOutput->parray);
		vOutput->parray = NULL;
	}

	vOutput->vt = VT_UI1 | VT_ARRAY;

	// It may be worthy to put a timeout value here but as is, this would allow a user to leave
	// a session open for days without timing out.
	WaitForSingleObject(m_hOutputMutex, INFINITE);

	// After this statement, m_pOutList is valid.
	if (m_pOutList == NULL)
	{
		// Nothing in the list so release the mutex and wait for something to be added to the list.
		// Release the mutex
		ReleaseMutex(m_hOutputMutex);
		// Wait for the event signalling new data has been added to the list
		WaitForSingleObject(m_hOutputEvent, INFINITE);
		// Wait for the mutex to syncronize access to the list
		WaitForSingleObject(m_hOutputMutex, INFINITE);
	}

	bound.lLbound = 0;
	bound.cElements = m_pOutList->size;

	// Create an array to return the data in.
	vOutput->parray = SafeArrayCreate(VT_UI1, 1, &bound);

	// Copy the data in the list to the array
	if (m_pOutList->size > 0)
	{
		SafeArrayAccessData(vOutput->parray, &pBuf);
		memcpy(pBuf, m_pOutList->data, m_pOutList->size);
		SafeArrayUnaccessData(vOutput->parray);
	}

	// Set who the chunk of data came from
	*from = m_pOutList->from;

	// Remove the node that has just been copied and signal whether there is potentially more data to come.
	node = m_pOutList;

	m_pOutList = m_pOutList->next;
	if (m_pOutList == NULL)
	{
		m_pOutListTail = NULL;
		ResetEvent(m_hOutputEvent);
	}

	if (node->size > 0)
	{
		delete node->data;
		*more = 1;
	}
	else
	{
		m_nOutputStreamsRemaining--;
		if (m_nOutputStreamsRemaining)
			*more = 1;
		else
			*more = 0;
	}
	delete node;

	ReleaseMutex(m_hOutputMutex);

	}catch(...){
		*error = 1;
		SysReAllocString(berror_msg, L"GetJobOutput:Exception thrown");
		LogWMsg(L"Exception thrown in GetInteractiveOutput.\n");
	}
	return S_OK;
}

// Function name	: CMPIJob::Abort
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::Abort(long *error, BSTR *berror_msg)
{
	// This is not a thread safe sequence because a process
	// could be launched between the Wait call and the PulseEvent call
	if (WaitForSingleObject(m_hAbortEvent, 0) != WAIT_OBJECT_0)
		PulseEvent(m_hAbortEvent);

	if ((m_nNumThreads > 0) && m_bLaunchThreadStarted)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (int i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = NULL;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_ABORT;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		//WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);
		if (WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, 10000) == WAIT_OBJECT_0)
		{
			for (i=0; i<m_nNumThreads; i++)
			{
				if (m_pThreadArgs[i].nResult != 0)
				{
					CloseHandle(m_hThreads[i]);
					m_hThreads[i] = NULL;
					if (*error == 0)
					{
						WCHAR wError[1024];
						*error = m_pThreadArgs[i].nResult;
						if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
							swprintf(wError, L"Abort:%s", m_pThreadArgs[i].wErrorMsg);
						else
							Translate_Error(*error, wError, L"Abort:ExtendJobThread failed ");
						SysReAllocString(berror_msg, wError);
						LogWMsg(wError);
					}
				}
			}
		}

		delete hEvents;
	}

	if (m_bLaunchThreadStarted) // Don't wait if the launch thread hasn't been started because it is the one which sets m_hAbortEvent
		WaitForSingleObject(m_hAbortEvent, 5000);

	DestroyJob(error, berror_msg);

	return S_OK;
}

// Function name	: CMPIJob::SetWorkingDirectory
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bDirectory
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::SetWorkingDirectory(BSTR bDirectory, long *error, BSTR *berror_msg)
{
	int i;

	*error = 0;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"SetWorkingDirectory:SetWorkingDirectory cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	wcscpy(m_pszWorkingDir, bDirectory);

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = bDirectory;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_SETDIR;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					WCHAR wError[1024];
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wError, L"SetWorkingDirectory:%s", m_pThreadArgs[i].wErrorMsg);
					else
						Translate_Error(*error, wError, L"SetWorkingDirectory:ExtendJobThread failed ");
					SysReAllocString(berror_msg, wError);
					LogWMsg(wError);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

// Function name	: CMPIJob::PutJobInput
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT vInput
// Argument         : long to
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPIJob::PutJobInput(VARIANT vInput, long to, long *error, BSTR *berror_msg)
{
	DWORD num_elements;
	LPVOID pBuf;
	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"PutJobInput:PutJobInput cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	if (vInput.parray != NULL)
	{
		num_elements = vInput.parray->rgsabound[0].cElements;
		if (num_elements > 0)
		{
			// Add the data to the output queue
			InputNode *n = new InputNode;
			n->data = new unsigned char[num_elements];
			SafeArrayAccessData(vInput.parray, &pBuf);
			memcpy(n->data, pBuf, num_elements);
			SafeArrayUnaccessData(vInput.parray);
			n->size = num_elements;
			n->next = NULL;
			if (WaitForSingleObject(m_hInputMutex, 10000) == WAIT_OBJECT_0)
			{
				if (m_pInListTail == NULL)
				{
					m_pInList = m_pInListTail = n;
				}
				else
				{
					m_pInListTail->next = n;
					m_pInListTail = n;
				}
				SetEvent(m_hInputEvent);
				ReleaseMutex(m_hInputMutex);
			}
			else
			{
				delete n->data;
				delete n;
			}
			
			// Destroy the array.
			//SafeArrayDestroy(vInput.parray);
		}
	}
	
	// TODO:
	// Signal the threads to pass the data along
	// ...
	// (For now, only the root process gets stdin redirected to it)

	return S_OK;
}

STDMETHODIMP CMPIJob::SetEnvironmentVariables(BSTR bEnv, long *error, BSTR *berror_msg)
{
	int i;

	*error = 0;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"SetEnvironmentVariables:SetEnvironmentVariables cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	wcscpy(m_pszEnv, bEnv);
	DLogWMsg(L"SetEnvironmentVariables: %s\n", bEnv);

	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].pCommandData = bEnv;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_SETENV;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					WCHAR wBuffer[1024];
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wBuffer, L"SetEnvironmentVariables:%s", m_pThreadArgs[i].wErrorMsg);
					else
					{
						Translate_Error(m_pThreadArgs[i].nResult, L"SetEnvironmentVariables failed ");
						wcscpy(wBuffer, m_pThreadArgs[i].wErrorMsg);
					}
					SysReAllocString(berror_msg, wBuffer);
					LogWMsg(wBuffer);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}

STDMETHODIMP CMPIJob::SendBreak(long *error, BSTR *berror_msg)
{
	int i;

	*error = 0;

	if (m_bInconsistentState)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"SendBreak:SendBreak cannot be called due to an error which occurred in CreateJob");
		return S_OK;
	}

	// Signal all the PutInputThreads to call pLaunch->SendBreak()
	PulseEvent(m_hBreakEvent);

	// Signal all the ExtendJobThreads to call SendBreak
	if (m_nNumThreads > 0)
	{
		HANDLE *hEvents = new HANDLE[m_nNumThreads];

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_hThreads[i] != NULL)
			{
				ResetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
				hEvents[i] = m_pThreadArgs[i].hCommandCompletedEvent;
				m_pThreadArgs[i].nCommand = MPIJOB_THRD_CMD_SENDBREAK;
				SetEvent(m_pThreadArgs[i].hCommandEvent);
			}
			else
			{
				m_pThreadArgs[i].nResult = E_FAIL;
				SetEvent(m_pThreadArgs[i].hCommandCompletedEvent);
			}
		}

		// TODO: Insert reasonable timeout
		WaitForMultipleObjects(m_nNumThreads, hEvents, TRUE, INFINITE);

		for (i=0; i<m_nNumThreads; i++)
		{
			if (m_pThreadArgs[i].nResult != 0)
			{
				CloseHandle(m_hThreads[i]);
				m_hThreads[i] = NULL;
				if (*error == 0)
				{
					WCHAR wBuffer[1024];
					*error = m_pThreadArgs[i].nResult;
					if (wcslen(m_pThreadArgs[i].wErrorMsg) > 0)
						swprintf(wBuffer, L"SendBreak:%s", m_pThreadArgs[i].wErrorMsg);
					else
					{
						Translate_Error(m_pThreadArgs[i].nResult, L"SendBreak failed ");
						wcscpy(wBuffer, m_pThreadArgs[i].wErrorMsg);
					}
					SysReAllocString(berror_msg, wBuffer);
					LogWMsg(wBuffer);
				}
			}
		}

		delete hEvents;
	}

	return S_OK;
}
