// MPILaunch.cpp : Implementation of CMPILaunch
#include "stdafx.h"
#include "MPILaunchServer.h"
#include "MPILaunch.h"
#include "..\Common\MPIJobLog.h"
#include "..\Common\Translate_Error.h"
#include <stdio.h>
#include "AccessDesktop.h"
#include <time.h>

/////////////////////////////////////////////////////////////////////////////
// CMPILaunch

// Function name	: CMPILaunch::~CMPILaunch
// Description	    : 
// Return type		: 
CMPILaunch::~CMPILaunch()
{
	try{

	if (m_hProcess)
	{
		DWORD exitCode;
		GetExitCodeProcess(m_hProcess, &exitCode);
		if (exitCode == STILL_ACTIVE)
			TerminateProcess(m_hProcess, 0);
		CloseHandle(m_hProcess);
		//LogMsg(TEXT("process handle closed"));
	}
	//else LogMsg(TEXT("NULL process handle"));
	m_hProcess = NULL;

	if (m_hOutputThread != NULL)
	{
		TerminateThread(m_hOutputThread, 0);
		CloseHandle(m_hOutputThread);
		m_hOutputThread = NULL;
	}

	if (m_hOutputMutex)
		CloseHandle(m_hOutputMutex);
	m_hOutputMutex = NULL;

	if (m_hOutputEvent)
		CloseHandle(m_hOutputEvent);
	m_hOutputEvent = NULL;

	if (m_hInteractivePipeR)
		CloseHandle(m_hInteractivePipeR);
	m_hInteractivePipeR = NULL;

	if (m_hInteractivePipeW)
		CloseHandle(m_hInteractivePipeW);
	m_hInteractivePipeW = NULL;

	if (m_pOutList)
	{
		ChunkNode *n;

		while (m_pOutList != NULL)
		{
			n = m_pOutList;
			m_pOutList = m_pOutList->next;
			if (n->data != NULL)
				delete n->data;
			delete n;
		}
	}
	m_pOutList = NULL;
	m_pOutListTail = NULL;
	//LogMsg(TEXT("x\n"));
	} catch(...){
		LogMsg(TEXT("Exception thrown in CMPILaunch destructor.\n"));
	}
}

// Function name	: RedirectOutput
// Description	    : 
// Return type		: void 
// Argument         : CYLaunchCOM *pCom
void RedirectOutput(CMPILaunch *pCom)
{
	unsigned char buffer[1024];
	DWORD num_read = 1024;

	DLogMsg(TEXT("RedirectOutput thread started by CMPILaunch\n"));
	while (num_read > 0)
	{
		num_read = 1024;
		if (ReadFile(pCom->m_hInteractivePipeR, buffer, num_read, &num_read, NULL))
		{
			if (num_read > 0)
			{
				DLogMsg(TEXT("RedirectOutput: %d bytes read from pipe, about to add to list.\n"), num_read);
				// Insert a node in the list.
				ChunkNode *n = new ChunkNode;
				n->data = new char[num_read];
				memcpy(n->data, buffer, num_read);
				n->size = num_read;
				n->next = NULL;
				WaitForSingleObject(pCom->m_hOutputMutex, INFINITE);
				if (pCom->m_pOutListTail == NULL)
				{
					pCom->m_pOutList = pCom->m_pOutListTail = n;
				}
				else
				{
					pCom->m_pOutListTail->next = n;
					pCom->m_pOutListTail = n;
				}
				SetEvent(pCom->m_hOutputEvent);
				ReleaseMutex(pCom->m_hOutputMutex);
				DLogMsg(TEXT("RedirectOutput: data added to m_pOutList.\n"));
			}
			else
			{
				// ReadFile returned zero bytes so the pipes must have closed.
				DLogMsg(TEXT("RedirectOutput: zero bytes read from pipe.\n"));
				WaitForSingleObject(pCom->m_hOutputMutex, INFINITE);
				CloseHandle(pCom->m_hInteractivePipeR);
				CloseHandle(pCom->m_hInteractivePipeW);
				pCom->m_hInteractivePipeR = NULL;
				pCom->m_hInteractivePipeW = NULL;
				ReleaseMutex(pCom->m_hOutputMutex);
			}
		}
		else
		{
			// ReadFile failed so the process must have exited.
			DLogMsg(TEXT("RedirectOutput: ReadFile failed.\n"));
			CloseHandle(pCom->m_hInteractivePipeR);
			CloseHandle(pCom->m_hInteractivePipeW);
			pCom->m_hInteractivePipeR = NULL;
			pCom->m_hInteractivePipeW = NULL;
			break;
		}
	}

	// Insert a node indicating the end of the stream.
	DLogMsg(TEXT("RedirectOutput: inserting last node to signal no more data.\n"));
	ChunkNode *n = new ChunkNode;
	n->size = 0; // <---- 0 size indicates end of stream.
	n->data = NULL; // <---- NULL data indicates end of stream.
	n->next = NULL;
	GetExitCodeProcess(pCom->m_hProcess, &n->exitcode);
	WaitForSingleObject(pCom->m_hOutputMutex, INFINITE);
	if (pCom->m_pOutListTail == NULL)
		pCom->m_pOutList = pCom->m_pOutListTail = n;
	else
	{
		pCom->m_pOutListTail->next = n;
		pCom->m_pOutListTail = n;
	}
	SetEvent(pCom->m_hOutputEvent);

	// Close the thread handle because the thread is about to exit.
	// Other threads may access this thread handle but no other thread
	// is checking to see when this thread exits so I close the handle here.
	// These two lines of code are protected by m_hOutputMutex
	CloseHandle(pCom->m_hOutputThread);
	pCom->m_hOutputThread = NULL;

	ReleaseMutex(pCom->m_hOutputMutex);
}

// Function name	: SetEnvironmentVariables
// Description	    : 
// Return type		: void 
// Argument         : BSTR bEnv
void SetEnvironmentVariables(BSTR bEnv)
{
	TCHAR name[MAX_PATH]=_T(""), value[MAX_PATH]=_T("");
	WCHAR wName[MAX_PATH]=L"", wValue[MAX_PATH]=L"";
	WCHAR *pChar;

	pChar = wName;
	while (*bEnv != L'\0')
	{
		if (*bEnv == L'=')
		{
			*pChar = L'\0';
			pChar = wValue;
		}
		else
		if (*bEnv == L'|')
		{
			*pChar = L'\0';
			pChar = wName;
#ifdef UNICODE
			wcscpy(name, wName);
			wcscpy(value, wValue);
#else
			wcstombs(name, wName, wcslen(wName)+1);
			wcstombs(value, wValue, wcslen(wValue)+1);
#endif
			//LogWMsg(L"name: '%s', value: '%s'\n", wName, wValue);
			SetEnvironmentVariable(name, value);
		}
		else
		{
			*pChar = *bEnv;
			pChar++;
		}
		bEnv++;
	}
	*pChar = L'\0';
#ifdef UNICODE
	wcscpy(name, wName);
	wcscpy(value, wValue);
#else
	wcstombs(name, wName, wcslen(wName)+1);
	wcstombs(value, wValue, wcslen(wValue)+1);
#endif
	//LogWMsg(L"name: '%s', value: '%s'\n", wName, wValue);
	SetEnvironmentVariable(name, value);
}

// Function name	: RemoveEnvironmentVariables
// Description	    : 
// Return type		: void 
// Argument         : BSTR bEnv
void RemoveEnvironmentVariables(BSTR bEnv)
{
	TCHAR name[MAX_PATH]=_T("");
	WCHAR wName[MAX_PATH]=L"", wValue[MAX_PATH]=L"";
	WCHAR *pChar;

	pChar = wName;
	while (*bEnv != L'\0')
	{
		if (*bEnv == L'=')
		{
			*pChar = L'\0';
			pChar = wValue;
		}
		else
		if (*bEnv == L'|')
		{
			*pChar = L'\0';
			pChar = wName;
#ifdef UNICODE
			wcscpy(name, wName);
#else
			wcstombs(name, wName, wcslen(wName)+1);
#endif
			//LogWMsg(L"name: '%s', value: '%s'\n", wName, wValue);
			SetEnvironmentVariable(name, NULL);
		}
		else
		{
			*pChar = *bEnv;
			pChar++;
		}
		bEnv++;
	}
	*pChar = L'\0';
#ifdef UNICODE
	wcscpy(name, wName);
#else
	wcstombs(name, wName, wcslen(wName)+1);
#endif
	SetEnvironmentVariable(name, NULL);
}

void RemoveMPIEnvironmentVariables()
{
	SetEnvironmentVariable("MPICH_JOBID", NULL);
	SetEnvironmentVariable("MPICH_IPROC", NULL);
	SetEnvironmentVariable("MPICH_NPROC", NULL);
	SetEnvironmentVariable("MPICH_ROOTHOST", NULL);
	SetEnvironmentVariable("MPICH_ROOTPORT", NULL);
	SetEnvironmentVariable("MPICH_EXTRA", NULL);
}

// Function name	: ParseAccountDomain
// Description	    : 
// Return type		: void 
// Argument         : LPCWSTR bAccount
// Argument         : LPTSTR tAccount
// Argument         : LPTSTR tDomain
void ParseAccountDomain(LPCWSTR bAccount, LPTSTR tAccount, LPTSTR tDomain)
{
	WCHAR wAccount[100]=L"", wDomain[100]=L"", *pwCh2;
	LPCWSTR pwCh;

	pwCh = bAccount;
	pwCh2 = wDomain;
	while ((*pwCh != L'\\') && (*pwCh != L'\0'))
	{
		*pwCh2 = *pwCh;
		pwCh++;
		pwCh2++;
	}
	if (*pwCh == L'\\')
	{
		pwCh++;
		wcscpy(wAccount, pwCh);
		*pwCh2 = L'\0';
	}
	else
	{
		wcscpy(wAccount, bAccount);
		wDomain[0] = L'\0';
	}
#ifdef UNICODE
	wcscpy(tAccount, wAccount);
	wcscpy(tDomain, wDomain);
#else
	wcstombs(tAccount, wAccount, wcslen(wAccount)+1);
	wcstombs(tDomain, wDomain, wcslen(wDomain)+1);
#endif
}

void ParseAccountDomainW(LPCWSTR bAccount, LPWSTR wAccount, LPWSTR wDomain)
{
	WCHAR *pwCh2;
	LPCWSTR pwCh;

	wAccount[0] = L'\0';
	wDomain[0] = L'\0';

	pwCh = bAccount;
	pwCh2 = wDomain;
	while ((*pwCh != L'\\') && (*pwCh != L'\0'))
	{
		*pwCh2 = *pwCh;
		pwCh++;
		pwCh2++;
	}
	if (*pwCh == L'\\')
	{
		pwCh++;
		wcscpy(wAccount, pwCh);
		*pwCh2 = L'\0';
	}
	else
	{
		wcscpy(wAccount, bAccount);
		wDomain[0] = L'\0';
	}
}

// Function name	: CMPILaunch::LaunchInteractiveMPIProcess
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bCmdLine
// Argument         : BSTR bDir
// Argument         : BSTR bEnv
// Argument         : BSTR bAccount
// Argument         : BSTR bPassword
// Argument         : long *pid
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::LaunchInteractiveMPIProcess(BSTR bCmdLine, BSTR bDir, BSTR bEnv, BSTR bAccount, BSTR bPassword, long *pid, long *error, BSTR *berror_msg)
{
	WCHAR error_msg[256];
	BOOL bSuccess = FALSE;
	HANDLE hStdin, hStdout, hStderr;
	HANDLE hPipeR=NULL, hPipeW=NULL;
	HANDLE thPipeR=NULL, thPipeW=NULL;
	HANDLE hUser = NULL;
	STARTUPINFO saInfo;
	PROCESS_INFORMATION psInfo;
	void *pEnv=NULL;
	TCHAR tAccount[256], tPassword[256], tCmdLine[MAX_PATH], tDomain[256], *psztDomain;
	TCHAR tSavedPath[MAX_PATH] = TEXT(".");
	DWORD nMaxPipeBuffer = 1;

	try{
	DLogMsg(TEXT("LaunchInteractiveProcess called: %u\n"), this);
	DLogWMsg(L"\n     Launching:\n        %s\n        %s\n        %s\n\n", bCmdLine, bEnv, bAccount);

	ParseAccountDomain(bAccount, tAccount, tDomain);
	if (_tcslen(tDomain) < 1)
		psztDomain = NULL;
	else
		psztDomain = tDomain;
#ifdef UNICODE
	wcscpy(tPassword, bPassword);
	wcscpy(tCmdLine, bCmdLine);
	//swprintf(tCmdLine, L"cmd.exe /c %s", bCmdLine);
#else
	wcstombs(tPassword, bPassword, wcslen(bPassword)+1);
	wcstombs(tCmdLine, bCmdLine, wcslen(bCmdLine)+1);
	//char sTempBuffer[256];
	//wcstombs(sTempBuffer, bCmdLine, wcslen(bCmdLine)+1);
	//sprintf(tCmdLine, "cmd.exe /c %s", sTempBuffer);
#endif

	// Launching of the client processes must be synchronized because
	// stdin,out,err are redirected for the entire process, not just this thread.
	if (WaitForSingleObject(g_hLaunchSyncMutex, 5000) == WAIT_TIMEOUT)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchInteractiveMPIProcess: Timeout while waiting for syncronization object.\n");
		LogMsg(TEXT("LaunchInteractiveProcess: Timeout while waiting for syncronization object.\n"));
		return S_OK;
	}

	// Don't handle errors, just let the process die.
	// In the future this will be configurable to allow various debugging options.
	SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX);

	// Save stdin, stdout, and stderr
	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	hStderr = GetStdHandle(STD_ERROR_HANDLE);
	if (hStdin == INVALID_HANDLE_VALUE || hStdout == INVALID_HANDLE_VALUE  || hStderr == INVALID_HANDLE_VALUE)
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"LaunchInteractiveMPIProcess: Unable to get standard handles.\n");
		ReleaseMutex(g_hLaunchSyncMutex);
		return S_OK;
	}

	// Set the security attributes to allow handles to be inherited
	SECURITY_ATTRIBUTES saAttr;
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.lpSecurityDescriptor = NULL;
	saAttr.bInheritHandle = TRUE;

	// Create pipes for stdin, stdout, and stderr
	if (!CreatePipe(&hPipeR, &thPipeW, &saAttr, 0))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:CreatePipe failed ");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}
	if (!CreatePipe(&thPipeR, &hPipeW, &saAttr, 0))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:CreatePipe failed");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}

	// Make the ends of the pipes that this process will use not inheritable
	if (!DuplicateHandle(GetCurrentProcess(), thPipeW, 
		GetCurrentProcess(), &m_hInteractivePipeW, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:DuplicateHandle(WritePipe) failed ");
		SysReAllocString(berror_msg, error_msg);
		thPipeW = NULL;
		goto CLEANUP;
	}
	thPipeW = NULL;

	if (!DuplicateHandle(GetCurrentProcess(), thPipeR, 
		GetCurrentProcess(), &m_hInteractivePipeR, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:DuplicateHanlde(ReadPipe) failed ");
		SysReAllocString(berror_msg, error_msg);
		thPipeR = NULL;
		goto CLEANUP;
	}
	thPipeR = NULL;
		
	// Set stdin, stdout, and stderr to the ends of the pipe the created process will use
	if (!SetStdHandle(STD_INPUT_HANDLE, hPipeR))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(Input) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}
	if (!SetStdHandle(STD_OUTPUT_HANDLE, hPipeW))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(Output) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto RESTORE_CLEANUP;
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, hPipeW)) // stderr and stdout will share a pipe for now
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(Error) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto RESTORE_CLEANUP;
	}

	// Create the process
	memset(&saInfo, 0, sizeof(STARTUPINFO));
	saInfo.cb = sizeof(STARTUPINFO);
	saInfo.hStdError = hPipeW;
	saInfo.hStdInput = hPipeR;
	saInfo.hStdOutput = hPipeW;
	saInfo.dwFlags = STARTF_USESTDHANDLES;
	//saInfo.lpDesktop = TEXT("WinSta0\\Default");
	if (m_bLaunchOnDesktop)
	{
		saInfo.lpDesktop  = TEXT("WinSta0\\Default");
		//saInfo.wShowWindow = SW_SHOW;
	}

	SetEnvironmentVariables(bEnv);
	pEnv = GetEnvironmentStrings();

	GetCurrentDirectory(MAX_PATH, tSavedPath);
	if (!SetCurrentDirectoryW(bDir))
	{
		int terror = GetLastError();
		char terror_msg[256];
		Translate_Error(terror, terror_msg, "LaunchInteractiveMPIProcess:SetCurrentDirectory failed ");
		LogMsg(terror_msg);
	}
	//LogWMsg(L"logging on user: %s\n", bAccount);
	if (LogonUser(
		tAccount,
		psztDomain, 
		tPassword,
		LOGON32_LOGON_INTERACTIVE, 
		LOGON32_PROVIDER_DEFAULT, 
		&hUser))
	{
		//LogMsg(TEXT("impersonating user\n"));
		if (ImpersonateLoggedOnUser(hUser))
		{
			//LogMsg(TEXT("LaunchInteractiveProcess: about to launch %s.\n"), tCmdLine);
			if (CreateProcessAsUser(
				hUser,
				NULL,
				tCmdLine,
				NULL, NULL, TRUE,
				//DETACHED_PROCESS | IDLE_PRIORITY_CLASS, 
				//CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS,
				CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP,
				//DETACHED_PROCESS | IDLE_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP,
				//CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS | CREATE_SUSPENDED, 
				pEnv,
				NULL,
				&saInfo, &psInfo))
			{
				m_hProcess = psInfo.hProcess;
				//ResumeThread(psInfo.hThread);
				CloseHandle(psInfo.hThread);
				LogMsg(TEXT("LaunchInteractiveMPIProcess: launched %s\n"), tCmdLine);
				bSuccess = TRUE;
				*pid = psInfo.dwProcessId;
				m_dwProcessId = psInfo.dwProcessId;
				SysReAllocString(berror_msg, L"success");
				*error = 0;
			}
			else
			{
				*error = GetLastError();
				Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:CreateProcessAsUser failed: ");
				SysReAllocString(berror_msg, error_msg);
				LogWMsg(L"LaunchInteractiveProcess: CreateProcessAsUser failed: %d, %s\n", *error, error_msg);
			}
			RevertToSelf();
		}
		else
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:ImpersonateLoggedOnUser failed ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"LaunchInteractiveProcess: ImpersonateLoggedOnUser failed: %d, %s\n", *error, error_msg);
		}
	}
	else
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:LogonUser failed ");
		SysReAllocString(berror_msg, error_msg);
		LogWMsg(L"LaunchInteractiveProcess: LogonUser failed: %d, %s\n", *error, error_msg);
	}
	CloseHandle(hUser);

	FreeEnvironmentStrings((TCHAR*)pEnv);
	SetCurrentDirectory(tSavedPath);
	RemoveEnvironmentVariables(bEnv);

RESTORE_CLEANUP:
	// Restore stdin, stdout, stderr
	if (!SetStdHandle(STD_INPUT_HANDLE, hStdin))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(restore Input) failed ");
		SysReAllocString(berror_msg, error_msg);
	}
	if (!SetStdHandle(STD_OUTPUT_HANDLE, hStdout))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(restore Output) failed ");
		SysReAllocString(berror_msg, error_msg);
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, hStderr))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveMPIProcess:SetStdHandle(restore Error) failed ");
		SysReAllocString(berror_msg, error_msg);
	}

	if (bSuccess)
	{
		// start thread to monitor output of pipe
		DWORD dwThreadID;
		m_hOutputThread = CreateThread(
			NULL, 0, 
			(LPTHREAD_START_ROUTINE)RedirectOutput,
			this,
			0,
			&dwThreadID);
		
		if (m_hOutputThread == NULL)
			SysReAllocString(berror_msg, L"Unable to create a thread to redirect output.\n");
	}

CLEANUP:
	ReleaseMutex(g_hLaunchSyncMutex);
	CloseHandle(hPipeR);
	CloseHandle(hPipeW);

	}catch(...){
		*error = 1;
		ReleaseMutex(g_hLaunchSyncMutex);
		SysReAllocString(berror_msg, L"LaunchInteractiveMPIProcess:Exception thrown");
		LogWMsg(L"Exception thrown in LaunchInteractiveMPIProcess.\n");
	}
	return S_OK;
}

// Function name	: CMPILaunch::GetInteractiveOutput
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT *vOutput
// Argument         : long *more
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::GetInteractiveOutput(VARIANT *vOutput, long *more, long *error, BSTR *berror_msg)
{
	SAFEARRAYBOUND bound;
	void *pBuf;
	ChunkNode *node;

	try{
	VariantClear(vOutput);

	vOutput->vt = VT_UI1 | VT_ARRAY;

	// It may be worthy to put a timeout value here but as is, this would allow a user to leave
	// a session open for days without timing out.
	WaitForSingleObject(m_hOutputMutex, INFINITE);

	// After this block, m_pOutList is valid.
	if (m_pOutList == NULL)
	{
		// Nothing in the list so release the mutex and wait for something to be added to the list.
		// Release the mutex
		ReleaseMutex(m_hOutputMutex);
		// Wait for the event signalling new data has been added to the list
		WaitForSingleObject(m_hOutputEvent, INFINITE);
		// Wait for the mutex to syncronize access to the list
		WaitForSingleObject(m_hOutputMutex, INFINITE);
	}

	bound.lLbound = 0;
	bound.cElements = m_pOutList->size;

	// Create an array to return the data in.
	vOutput->parray = SafeArrayCreate(VT_UI1, 1, &bound);

	// Copy the data in the list to the array
	if (m_pOutList->size > 0)
	{
		SafeArrayAccessData(vOutput->parray, &pBuf);
		memcpy(pBuf, m_pOutList->data, m_pOutList->size);
		SafeArrayUnaccessData(vOutput->parray);
	}

	// Remove the node that has just been copied and signal whether there is potentially more data to come.
	node = m_pOutList;

	m_pOutList = m_pOutList->next;
	if (m_pOutList == NULL)
	{
		m_pOutListTail = NULL;
		ResetEvent(m_hOutputEvent);
	}

	if (node->size > 0)
	{
		delete node->data;
		*more = 1;
	}
	else
	{
		*more = 0;
		//*error = node->exitcode;
	}

	delete node;

	ReleaseMutex(m_hOutputMutex);

	}catch(...){
		*error = 1;
		SysReAllocString(berror_msg, L"GetInteractiveOutput:Exception thrown");
		LogWMsg(L"Exception thrown in GetInteractiveOutput.\n");
	}
	return S_OK;
}

// Function name	: CMPILaunch::LaunchMPIProcess
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bCmdLine
// Argument         : BSTR bDir
// Argument         : BSTR bEnv
// Argument         : BSTR bAccount
// Argument         : BSTR bPassword
// Argument         : long *pid
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::LaunchMPIProcess(BSTR bCmdLine, BSTR bDir, BSTR bEnv, BSTR bAccount, BSTR bPassword, long *pid, long *error, BSTR *berror_msg)
{
	STARTUPINFOW sinfo;
	PROCESS_INFORMATION pinfo;
	DWORD length = 0;
	WCHAR error_msg[256];
	TCHAR tSavedPath[MAX_PATH] = TEXT(".");
	void *pEnv = NULL;
	WCHAR wAccount[256], wDomain[256], *pszwDomain;

	try{
	length = wcslen(bCmdLine) + 1;
	if (length == 1)
	{
		*error = 1;
		*pid = 0;
		SysReAllocString(berror_msg, L"Error: Command line empty.\n");
		LogWMsg(L"LaunchProcess: Error: Command line empty.\n");
		return S_OK;
	}
	
	ParseAccountDomainW(bAccount, wAccount, wDomain);
	if (wcslen(wDomain) < 1)
		pszwDomain = NULL;
	else
		pszwDomain = wDomain;

	GetStartupInfoW(&sinfo);

	if (WaitForSingleObject(g_hLaunchSyncMutex, 5000) == WAIT_TIMEOUT)
	{
		*error = 1;
		*pid = 0;
		SysReAllocString(berror_msg, L"LaunchMPIProcess: Thread timed out waiting for the syncronization object.\n");
		LogWMsg(L"LaunchProcess: Error: Thread timed out waiting for the syncronization object.\n");
		return S_OK;
	}

	SetEnvironmentVariables(bEnv);
	pEnv = GetEnvironmentStrings();

	GetCurrentDirectory(MAX_PATH, tSavedPath);
	SetCurrentDirectoryW(bDir);

	HANDLE hUser = NULL;
	if (LogonUserW(
		wAccount,
		pszwDomain, 
		bPassword,
		LOGON32_LOGON_INTERACTIVE,
		LOGON32_PROVIDER_DEFAULT, &hUser))
	{
		if (ImpersonateLoggedOnUser(hUser))
		{
			if (CreateProcessAsUserW(
				hUser,
				NULL,
				bCmdLine,
				NULL, NULL, TRUE,
				DETACHED_PROCESS | IDLE_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP, 
				pEnv,
				NULL,
				&sinfo, &pinfo))
			{
				m_hProcess = pinfo.hProcess;
				CloseHandle(pinfo.hThread);
				LogWMsg(L"LaunchMPIProcess: launched %s\n", bCmdLine);
				*pid = pinfo.dwProcessId;
				m_dwProcessId = pinfo.dwProcessId;
				SysReAllocString(berror_msg, L"success");
				*error = 0;
			}
			else
			{
				*error = GetLastError();
				*pid = 0;
				Translate_Error(*error, error_msg, L"LaunchMPIProcess:CreateProcessAsUser failed ");
				SysReAllocString(berror_msg, error_msg);
				LogWMsg(L"LaunchProcess: CreateProcess failed: %s\n", error_msg);
			}
			RevertToSelf();
		}
		else
		{
			*error = GetLastError();
			*pid = 0;
			Translate_Error(*error, error_msg, L"LaunchMPIProcess:ImpersonateLoggedOnUser failed ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"LaunchProcess: ImpersonateLoggedOnUser failed: %s\n", error_msg);
		}
	}
	else
	{
		*error = GetLastError();
		*pid = 0;
		Translate_Error(*error, error_msg, L"LaunchMPIProcess:LogonUser failed ");
		SysReAllocString(berror_msg, error_msg);
		LogWMsg(L"LaunchProcess: LogonUser failed: %s\n", error_msg);
	}
	CloseHandle(hUser);
	FreeEnvironmentStrings((TCHAR*)pEnv);
	SetCurrentDirectory(tSavedPath);
	
	ReleaseMutex(g_hLaunchSyncMutex);
	
	}catch(...){
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchMPIProcess: Exception thrown");
		LogWMsg(L"Exception thrown in LaunchProcess.\n");
	}
	return S_OK;
}

// Function name	: CMPILaunch::Abort
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::Abort(long *error, BSTR *berror_msg)
{
	if (m_hProcess != NULL)
	{
		*error = 1;
		if (GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, m_dwProcessId))
		{
			if (WaitForSingleObject(m_hProcess, 500) == WAIT_OBJECT_0)
				*error = 0;
		}
		if (*error)
		{
			if (TerminateProcess(m_hProcess, 1))
				*error = 0;
			else
			{
				*error = GetLastError();
				WCHAR error_msg[256];
				Translate_Error(*error, error_msg, L"Abort:TerminateProcess failed ");
				SysReAllocString(berror_msg, error_msg);
				LogWMsg(L"%d, %s", *error, error_msg);
			}
		}
	}
	return S_OK;
}

// Function name	: CMPILaunch::PutInteractiveInput
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : VARIANT vInput
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::PutInteractiveInput(VARIANT vInput, long *error, BSTR *berror_msg)
{
	LPVOID pBuf;
	DWORD size, num_written;

	if (vInput.vt == (VT_UI1 | VT_ARRAY))
	{
		size = vInput.parray->rgsabound->cElements;
		SafeArrayAccessData(vInput.parray, &pBuf);
		WriteFile(m_hInteractivePipeW, pBuf, size, &num_written, NULL);
		SafeArrayUnaccessData(vInput.parray);
	}

	return S_OK;
}

// Function name	: CMPILaunch::SendBreak
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::SendBreak(long *error, BSTR *berror_msg)
{
	if (m_hProcess != NULL)
	{
		if (!GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, m_dwProcessId))
		//if (!GenerateConsoleCtrlEvent(CTRL_C_EVENT, m_dwProcessId))
		{
			*error = GetLastError();
			WCHAR error_msg[256];
			Translate_Error(*error, error_msg, L"SendBreak:GenerateConsoleCtrlEvent failed ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"processID: %d, error: %d message: %s", m_dwProcessId, *error, error_msg);
		}
	}
	return S_OK;
}

// Function name	: CMPILaunch::LaunchInteractiveImpersontatedMPIProcess
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bCmdLine
// Argument         : BSTR bDir
// Argument         : BSTR bEnv
// Argument         : long *pid
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::LaunchInteractiveImpersontatedMPIProcess(BSTR bCmdLine, BSTR bDir, BSTR bEnv, long *pid, long *error, BSTR *berror_msg)
{
	WCHAR error_msg[256];
	BOOL bSuccess = FALSE;
	HANDLE hStdin, hStdout, hStderr;
	HANDLE hPipeR=NULL, hPipeW=NULL;
	HANDLE thPipeR=NULL, thPipeW=NULL;
	HANDLE hUser = NULL;
	STARTUPINFO saInfo;
	PROCESS_INFORMATION psInfo;
	void *pEnv=NULL;
	TCHAR tCmdLine[MAX_PATH];
	TCHAR tSavedPath[MAX_PATH] = TEXT(".");
	HANDLE hImpersonatedToken;
	HRESULT hr;

	try{
	DLogMsg(TEXT("LaunchInteractiveImpersontatedMPIProcess called: %u\n"), this);
	DLogWMsg(L"\n     Launching:\n        %s\n        %s\n\n", bCmdLine, bEnv);

#ifdef UNICODE
	wcscpy(tCmdLine, bCmdLine);
	//swprintf(tCmdLine, L"cmd.exe /c %s", bCmdLine);
#else
	wcstombs(tCmdLine, bCmdLine, wcslen(bCmdLine)+1);
	//char sTempBuffer[256];
	//wcstombs(sTempBuffer, bCmdLine, wcslen(bCmdLine)+1);
	//sprintf(tCmdLine, "cmd.exe /c %s", sTempBuffer);
#endif

	// Launching of the client processes must be synchronized because
	// stdin,out,err are redirected for the entire process, not just this thread.
	if (WaitForSingleObject(g_hLaunchSyncMutex, 5000) == WAIT_TIMEOUT)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"LaunchInteractiveImpersontatedMPIProcess: Timeout while waiting for syncronization object.\n");
		LogMsg(TEXT("LaunchInteractiveImpersontatedMPIProcess: Timeout while waiting for syncronization object.\n"));
		return S_OK;
	}

	// Don't handle errors, just let the process die.
	// In the future this will be configurable to allow various debugging options.
	SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX);

	// Save stdin, stdout, and stderr
	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	hStderr = GetStdHandle(STD_ERROR_HANDLE);
	if (hStdin == INVALID_HANDLE_VALUE || hStdout == INVALID_HANDLE_VALUE  || hStderr == INVALID_HANDLE_VALUE)
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"LaunchInteractiveImpersontatedMPIProcess: Unable to get standard handles.\n");
		ReleaseMutex(g_hLaunchSyncMutex);
		return S_OK;
	}

	// Set the security attributes to allow handles to be inherited
	SECURITY_ATTRIBUTES saAttr;
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.lpSecurityDescriptor = NULL;
	saAttr.bInheritHandle = TRUE;

	// Create pipes for stdin, stdout, and stderr
	if (!CreatePipe(&hPipeR, &thPipeW, &saAttr, 0))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:CreatePipe failed ");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}
	if (!CreatePipe(&thPipeR, &hPipeW, &saAttr, 0))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:CreatePipe failed");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}

	// Make the ends of the pipes that this process will use not inheritable
	if (!DuplicateHandle(GetCurrentProcess(), thPipeW, 
		GetCurrentProcess(), &m_hInteractivePipeW, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:DuplicateHandle(WritePipe) failed ");
		SysReAllocString(berror_msg, error_msg);
		thPipeW = NULL;
		goto CLEANUP;
	}
	thPipeW = NULL;

	if (!DuplicateHandle(GetCurrentProcess(), thPipeR, 
		GetCurrentProcess(), &m_hInteractivePipeR, 
		0, FALSE, DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:DuplicateHanlde(ReadPipe) failed ");
		SysReAllocString(berror_msg, error_msg);
		thPipeR = NULL;
		goto CLEANUP;
	}
	thPipeR = NULL;
		
	// Set stdin, stdout, and stderr to the ends of the pipe the created process will use
	if (!SetStdHandle(STD_INPUT_HANDLE, hPipeR))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(Input) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto CLEANUP;
	}
	if (!SetStdHandle(STD_OUTPUT_HANDLE, hPipeW))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(Output) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto RESTORE_CLEANUP;
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, hPipeW)) // stderr and stdout will share a pipe for now
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(Error) failed ");
		SysReAllocString(berror_msg, error_msg);
		goto RESTORE_CLEANUP;
	}

	// Set up the STARTINFO structure
	memset(&saInfo, 0, sizeof(STARTUPINFO));
	saInfo.cb         = sizeof(STARTUPINFO);
	saInfo.hStdInput  = hPipeR;
	saInfo.hStdOutput = hPipeW;
	saInfo.hStdError  = hPipeW;
	saInfo.dwFlags    = STARTF_USESTDHANDLES;
	if (m_bLaunchOnDesktop)
		saInfo.lpDesktop  = TEXT("WinSta0\\Default");

	// Set the environment variables and attempt to change into the directory passed into the function
	SetEnvironmentVariables(bEnv);
	pEnv = GetEnvironmentStrings();

	GetCurrentDirectory(MAX_PATH, tSavedPath);
	if (!SetCurrentDirectoryW(bDir))
	{
		int terror = GetLastError();
		char terror_msg[256];
		Translate_Error(terror, terror_msg, "LaunchInteractiveImpersontatedMPIProcess:SetCurrentDirectory failed ");
		LogMsg(terror_msg);
	}

	// Impersonate the client and get a user token to be used in CreateProcessAsUser
	hr = CoImpersonateClient();
	if (FAILED(hr))
		LogMsg(TEXT("LaunchInteractiveImpersontatedMPIProcess:CoImpersonateClient failed - launching process with process token"));
	//if (!OpenThreadToken(GetCurrentThread(), TOKEN_ALL_ACCESS, TRUE, &hImpersonatedToken))
	if (!OpenThreadToken(GetCurrentThread(), MAXIMUM_ALLOWED, TRUE, &hImpersonatedToken))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:OpenThreadToken failed: ");
		SysReAllocString(berror_msg, error_msg);
		LogWMsg(L"LaunchInteractiveImpersontatedMPIProcess:OpenThreadToken failed: %d, %s\n", *error, error_msg);
		return S_OK;
	}
	CoRevertToSelf();
	//if (!DuplicateTokenEx(hImpersonatedToken, TOKEN_ALL_ACCESS, NULL, SecurityImpersonation, TokenPrimary, &hUser))
	if (!DuplicateTokenEx(hImpersonatedToken, MAXIMUM_ALLOWED, NULL, SecurityImpersonation, TokenPrimary, &hUser))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:DuplicateTokenEx failed: ");
		SysReAllocString(berror_msg, error_msg);
		LogWMsg(L"LaunchInteractiveImpersontatedMPIProcess:DuplicateTokenEx failed: %d, %s\n", *error, error_msg);
		return S_OK;
	}

	// Create the process

	//LogMsg(TEXT("impersonating user\n"));
	if (ImpersonateLoggedOnUser(hUser))
	{
		//LogMsg(TEXT("LaunchInteractiveProcess: about to launch %s.\n"), tCmdLine);
		if (CreateProcessAsUser(
			hUser,
			NULL,
			tCmdLine,
			NULL, NULL, TRUE,
			//DETACHED_PROCESS | IDLE_PRIORITY_CLASS, 
			//CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS,
			CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP,
			//DETACHED_PROCESS | IDLE_PRIORITY_CLASS | CREATE_NEW_PROCESS_GROUP,
			//CREATE_NO_WINDOW | IDLE_PRIORITY_CLASS | CREATE_SUSPENDED, 
			pEnv,
			NULL,
			&saInfo, &psInfo))
		{
			m_hProcess = psInfo.hProcess;
			//ResumeThread(psInfo.hThread);
			CloseHandle(psInfo.hThread);
			LogMsg(TEXT("LaunchInteractiveImpersonatedMPIProcess: launched %s"), tCmdLine);
			bSuccess = TRUE;
			*pid = psInfo.dwProcessId;
			m_dwProcessId = psInfo.dwProcessId;
			SysReAllocString(berror_msg, L"success");
			*error = 0;
		}
		else
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:CreateProcessAsUser failed: ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"LaunchInteractiveImpersontatedMPIProcess: CreateProcessAsUser failed: %d, %s\n", *error, error_msg);
		}
		RevertToSelf();
	}
	else
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:ImpersonateLoggedOnUser failed ");
		SysReAllocString(berror_msg, error_msg);
		LogWMsg(L"LaunchInteractiveImpersontatedMPIProcess: ImpersonateLoggedOnUser failed: %d, %s\n", *error, error_msg);
	}
	CloseHandle(hUser);

	FreeEnvironmentStrings((TCHAR*)pEnv);
	SetCurrentDirectory(tSavedPath);
	RemoveEnvironmentVariables(bEnv);

RESTORE_CLEANUP:
	// Restore stdin, stdout, stderr
	if (!SetStdHandle(STD_INPUT_HANDLE, hStdin))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(restore Input) failed ");
		SysReAllocString(berror_msg, error_msg);
	}
	if (!SetStdHandle(STD_OUTPUT_HANDLE, hStdout))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(restore Output) failed ");
		SysReAllocString(berror_msg, error_msg);
	}
	if (!SetStdHandle(STD_ERROR_HANDLE, hStderr))
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"LaunchInteractiveImpersontatedMPIProcess:SetStdHandle(restore Error) failed ");
		SysReAllocString(berror_msg, error_msg);
	}

	if (bSuccess)
	{
		// start thread to monitor output of pipe
		DWORD dwThreadID;
		m_hOutputThread = CreateThread(
			NULL, 0, 
			(LPTHREAD_START_ROUTINE)RedirectOutput,
			this,
			0,
			&dwThreadID);
		
		if (m_hOutputThread == NULL)
			SysReAllocString(berror_msg, L"Unable to create a thread to redirect output.\n");
	}

CLEANUP:
	ReleaseMutex(g_hLaunchSyncMutex);
	CloseHandle(hPipeR);
	CloseHandle(hPipeW);

	}catch(...){
		*error = 1;
		ReleaseMutex(g_hLaunchSyncMutex);
		SysReAllocString(berror_msg, L"LaunchInteractiveImpersontatedMPIProcess:Exception thrown");
		LogWMsg(L"Exception thrown in LaunchInteractiveImpersontatedMPIProcess");
	}
	return S_OK;
}

// Function name	: CMPILaunch::GetPortFromFile
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bFilename
// Argument         : long *nPort
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::GetPortFromFile(BSTR bFilename, long *nPort, long *error, BSTR *berror_msg)
{
	WCHAR error_msg[256];

	HANDLE hFile = CreateFileW(bFilename, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		*error = GetLastError();
		Translate_Error(*error, error_msg, L"GetPortFromFile:CreateFile failed ");
		LogWMsg(error_msg);
		LogWMsg(bFilename);
		SysReAllocString(berror_msg, error_msg);
		return S_OK;
	}
				
	DWORD num_read = 0;
	TCHAR pBuffer[100] = _T("");
	LPTSTR pChar = pBuffer;
	clock_t cStart = clock();
	while (true)
	{
		num_read = 0;
		if (!ReadFile(hFile, pChar, 100, &num_read, NULL))
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"GetPortFromFile:ReadFile failed ");
			LogWMsg(error_msg);
			SysReAllocString(berror_msg, error_msg);
			CloseHandle(hFile);
			DeleteFileW(bFilename);
			return S_OK;
		}
		if (num_read == 0)
		{
			if (clock() - cStart > 10 * CLOCKS_PER_SEC)
			{
				DWORD dwExitCode;
				if (GetExitCodeProcess(m_hProcess, &dwExitCode))
				{
					if (dwExitCode != STILL_ACTIVE)
					{
						LogMsg(TEXT("GetPortFromFile:Process has exited without writing the port number to a file. Exit code: %d"), dwExitCode);
						swprintf(error_msg, L"GetPortFromFile:Process has exited, no port number in the file.\nProcess exit code: %d", dwExitCode);
						*error = 1;
						SysReAllocString(berror_msg, error_msg);
						CloseHandle(hFile);
						DeleteFileW(bFilename);
						return S_OK;
					}
				}
				LogWMsg(L"GetPortFromFile:Wait for process 0 to write port to temporary file timed out: '%s'\n", bFilename);
				swprintf(error_msg, L"GetPortFromFile:Wait for process 0 to write port to temporary file timed out: '%s'", bFilename);
				*error = 1;
				SysReAllocString(berror_msg, error_msg);
				CloseHandle(hFile);
				DeleteFileW(bFilename);
				return S_OK;
			}
			Sleep(100);
		}
		else
		{
			for (unsigned int i=0; i<num_read; i++)
			{
				if (*pChar == _T('\n'))
					break;
				pChar ++;
			}
			if (*pChar == _T('\n'))
				break;
		}
	}
	CloseHandle(hFile);
	DeleteFileW(bFilename);

	*nPort = _ttoi(pBuffer);

	return S_OK;
}

// Function name	: CMPILaunch::CreateTempFile
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR *bFilename
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::CreateTempFile(BSTR *bFilename, long *error, BSTR *berror_msg)
{
	WCHAR wTemp[MAX_PATH];

	if (GetTempFileNameW(L".", L"mpi", 0, wTemp) == 0)
	{
		*error = GetLastError();
		Translate_Error(*error, wTemp, L"CreateTempFile:GetTempFileName failed ");
		LogWMsg(wTemp);
		SysReAllocString(berror_msg, wTemp);
		return S_OK;
	}

	WCHAR wFullTemp[MAX_PATH], *namepart;
	GetFullPathNameW(wTemp, MAX_PATH, wFullTemp, &namepart);

	SysReAllocString(bFilename, wFullTemp);

	return S_OK;
}

// Function name	: CMPILaunch::GrantAccessToDesktop
// Description	    : 
// Return type		: STDMETHODIMP 
// Argument         : BSTR bAccount
// Argument         : BSTR bPassword
// Argument         : long *error
// Argument         : BSTR *berror_msg
STDMETHODIMP CMPILaunch::GrantAccessToDesktop(BSTR bAccount, BSTR bPassword, long *error, BSTR *berror_msg)
{
	WCHAR error_msg[255];
	HANDLE hUser = NULL;
	HANDLE hImpersonatedToken = NULL;
	HRESULT hr = S_OK;
	try{
	if (wcslen(bAccount))
	{
		TCHAR tAccount[MAX_PATH], tPassword[MAX_PATH], tDomain[MAX_PATH], *psztDomain;

		ParseAccountDomain(bAccount, tAccount, tDomain);
		if (_tcslen(tDomain) < 1)
			psztDomain = NULL;
		else
			psztDomain = tDomain;
#ifdef UNICODE
		wcscpy(tPassword, bPassword);
#else
		wcstombs(tPassword, bPassword, wcslen(bPassword)+1);
#endif
		if (!LogonUser(
			tAccount,
			psztDomain, 
			tPassword,
			LOGON32_LOGON_INTERACTIVE, 
			LOGON32_PROVIDER_DEFAULT, 
			&hUser))
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"GrantAccessToDesktop:LogonUser failed: ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"GrantAccessToDesktop:LogonUser failed: %d, %s\n", *error, error_msg);
			return S_OK;
		}
	}
	else
	{
		// Impersonate the client and get a user token
		hr = CoImpersonateClient();
		if (FAILED(hr))
			LogMsg(TEXT("GrantAccessToDesktop:CoImpersonateClient failed"));
		if (!OpenThreadToken(GetCurrentThread(), MAXIMUM_ALLOWED, TRUE, &hImpersonatedToken))
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"GrantAccessToDesktop:OpenThreadToken failed: ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"GrantAccessToDesktop:OpenThreadToken failed: %d, %s\n", *error, error_msg);
			return S_OK;
		}
		CoRevertToSelf();
		if (!DuplicateTokenEx(hImpersonatedToken, MAXIMUM_ALLOWED, NULL, SecurityImpersonation, TokenPrimary, &hUser))
		{
			*error = GetLastError();
			Translate_Error(*error, error_msg, L"GrantAccessToDesktop:DuplicateTokenEx failed: ");
			SysReAllocString(berror_msg, error_msg);
			LogWMsg(L"GrantAccessToDesktop:DuplicateTokenEx failed: %d, %s\n", *error, error_msg);
			return S_OK;
		}
		CloseHandle(hImpersonatedToken);
		hImpersonatedToken = NULL;
	}
	
	m_bLaunchOnDesktop = MyGrantAccessToDesktop(hUser);

	CloseHandle(hUser);
	hUser = NULL;
	}catch(...)
	{
		if (hUser != NULL)
			CloseHandle(hUser);
		if (hImpersonatedToken != NULL)
			CloseHandle(hImpersonatedToken);
		LogMsg(TEXT("Exception thrown in GrantAccessToDesktop"));
	}
	return S_OK;
}
