#include <stdio.h>
#include <tchar.h>
#include "..\Common\MPICH_pwd.h"
#include "..\Common\MPIJobDefs.h"
#include <conio.h>
// MPIJobServer must be compiled before the following files are generated
#include "..\MPIJobServer\MPIJobServer.h"
#include <objbase.h>
#include "..\MPIJobServer\MPIJobserver_i.c"
#include <time.h>
#include "..\Common\Translate_Error.h"
#include "localonly.h"
#include "GetOpt.h"

#define TRANSFER_BUFFER_SIZE 100000

// Global variables
long g_nHosts = 1;
VARIANT g_vHosts, g_vSMPInfo;
long g_nFirstSMPProcs = 1;
TCHAR g_pszAccount[100], g_pszPassword[100];
bool g_bPush = false, g_bNoMPI = false;
TCHAR g_pszExe[MAX_PATH] = _T(""), g_pszArgs[MAX_PATH] = _T(""), g_pszEnv[1024] = _T("");
TCHAR g_pszFirstHost[100] = _T("");
IMPIJob *g_pJob = NULL;
HANDLE g_hFinishedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

#define MULTI_COLOR_OUTPUT
 
#ifdef MULTI_COLOR_OUTPUT
#include <wincon.h>

#define NUM_OUTPUT_COLORS 32

// foregrounds
#define frgnd_RGB	FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY
#define frgnd_RG	FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY
#define frgnd_RB	FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY
#define frgnd_R		FOREGROUND_RED | FOREGROUND_INTENSITY
#define frgnd_GB	FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY
#define frgnd_G		FOREGROUND_GREEN | FOREGROUND_INTENSITY
#define frgnd_B		FOREGROUND_BLUE | FOREGROUND_INTENSITY
#define frgnd_rgb	FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE
#define frgnd_rg	FOREGROUND_RED | FOREGROUND_GREEN
#define frgnd_rb	FOREGROUND_RED | FOREGROUND_BLUE
#define frgnd_r		FOREGROUND_RED
#define frgnd_gb	FOREGROUND_GREEN | FOREGROUND_BLUE
#define frgnd_g		FOREGROUND_GREEN
#define frgnd_b		FOREGROUND_BLUE

// backgrounds
#define bkgnd_RGB	BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY
#define bkgnd_RG	BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_INTENSITY
#define bkgnd_RB	BACKGROUND_RED | BACKGROUND_BLUE | BACKGROUND_INTENSITY
#define bkgnd_R		BACKGROUND_RED | BACKGROUND_INTENSITY
#define bkgnd_GB	BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY
#define bkgnd_G		BACKGROUND_GREEN | BACKGROUND_INTENSITY
#define bkgnd_B		BACKGROUND_BLUE | BACKGROUND_INTENSITY
#define bkgnd_rgb	BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE
#define bkgnd_rg	BACKGROUND_RED | BACKGROUND_GREEN
#define bkgnd_rb	BACKGROUND_RED | BACKGROUND_BLUE
#define bkgnd_r		BACKGROUND_RED
#define bkgnd_gb	BACKGROUND_GREEN | BACKGROUND_BLUE
#define bkgnd_g		BACKGROUND_GREEN
#define bkgnd_b		BACKGROUND_BLUE

WORD aConsoleColorAttribute[NUM_OUTPUT_COLORS] = {
	frgnd_rgb, 
	frgnd_RG,
	frgnd_RB,
	frgnd_R,
	frgnd_GB,
	frgnd_G,
	frgnd_RGB,
	frgnd_RGB | bkgnd_rb,
	frgnd_RGB | bkgnd_r,
	frgnd_RGB | bkgnd_gb,
	frgnd_RGB | bkgnd_g,
	frgnd_RGB | bkgnd_b,
	frgnd_RG | bkgnd_rb,
	frgnd_RG | bkgnd_r,
	frgnd_RG | bkgnd_gb,
	frgnd_RG | bkgnd_g,
	frgnd_RG | bkgnd_b,
	frgnd_RB | bkgnd_rb,
	frgnd_RB | bkgnd_b,
	frgnd_R | bkgnd_r,
	frgnd_R | bkgnd_b,
	frgnd_GB | bkgnd_rb,
	frgnd_GB | bkgnd_r,
	frgnd_GB | bkgnd_gb,
	frgnd_GB | bkgnd_g,
	frgnd_GB | bkgnd_b,
	frgnd_G | bkgnd_r,
	frgnd_G | bkgnd_gb,
	frgnd_G | bkgnd_g,
	frgnd_G | bkgnd_b,
	frgnd_rb | bkgnd_gb,
	frgnd_r | bkgnd_gb
};
#endif

// Function name	: PrintError
// Description	    : 
// Return type		: void 
// Argument         : HRESULT hr
void PrintError(HRESULT hr)
{
	HLOCAL str;
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER,
		0,
		hr,
		MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
		(LPTSTR) &str,
		0,0);
	
	_tprintf(TEXT("error: %s\n"), str);
	LocalFree(str);
}

// Function name	: Connect
// Description	    : 
// Return type		: bool 
// Argument         : LPTSTR host
// Argument         : IMPIJob *&pJob
bool Connect(LPTSTR host, IMPIJob *&pJob)
{
	HRESULT hr;
	MULTI_QI qi;
	WCHAR wHost[100], localhost[100];
	COSERVERINFO server;
	DWORD length = 100;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	if (FAILED(hr))
	{
		_tprintf(TEXT("CoInitialize() failed.\n"));
		PrintError(hr);
		return false;
	}

	//*
	hr = CoInitializeSecurity(NULL, -1, NULL, NULL,
		RPC_C_AUTHN_LEVEL_CONNECT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
		//RPC_C_AUTHN_LEVEL_PKT, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE, NULL);
	if (FAILED(hr))
	{
		if (hr == RPC_E_TOO_LATE)
			printf("CoInitializeSecurity failed in Connect(MPIJob) because it has already been set.\n");
		else
		{
			char error_msg[256];
			Translate_HRError(hr, error_msg);
			printf("CoInitializeSecurity failed in Connect(MPIJob)\nError: %s", error_msg);
		}
	}
	//*/

	qi.pIID = &IID_IMPIJob;
	qi.pItf = NULL;

#ifdef UNICODE
	wcscpy(wHost, host);
#else
	//MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, host, -1, wHost, 100);
	mbstowcs(wHost, host, strlen(host)+1);
#endif

	server.dwReserved1 = 0;
	server.dwReserved2 = 0;
	server.pAuthInfo = NULL;
	server.pwszName = wHost;

	GetComputerNameW(localhost, &length);

	if (_wcsicmp(localhost, wHost) == 0)
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}
	else
	{
		hr = CoCreateInstanceEx(CLSID_MPIJob, NULL, CLSCTX_REMOTE_SERVER, &server, 1, &qi);
		if (FAILED(hr))
		{
			_tprintf(TEXT("Unable to connect to %s: "), host);
			PrintError(hr);
			return false;
		}
	}

	if (SUCCEEDED(qi.hr))
	{
		pJob = (IMPIJob*)qi.pItf;
		return true;
	}

	return false;
}

// Function name	: TransferFile
// Description	    : 
// Return type		: void 
// Argument         : IMPIJob *pJob
// Argument         : LPTSTR filename
void TransferFile(IMPIJob *pJob, LPTSTR filename)
{
	HRESULT hr;
	VARIANT v;
	BYTE *buffer = new BYTE[TRANSFER_BUFFER_SIZE];
	DWORD num_read=TRANSFER_BUFFER_SIZE, num_written=0;
	LPVOID v_buf = NULL;
	HANDLE hLocalFile = NULL;
	LONG error;
	BSTR berror_msg;
				
	v.vt = VT_UI1 | VT_ARRAY;
	SAFEARRAYBOUND bound;
	bound.lLbound = 0;
				
	hLocalFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	
	if (hLocalFile != INVALID_HANDLE_VALUE)
	{
		berror_msg = SysAllocString(L"");
		error = 0;

		hr = g_pJob->StartExeTransfer(&error, &berror_msg);
		if (FAILED(hr))
		{
			_tprintf(TEXT("StartExeTransfer failed: "));
			PrintError(hr);
			pJob->DestroyJob(&error, &berror_msg);
			pJob->Release();
			ExitProcess(1);
		}
		if (error)
		{
			if (wcslen(berror_msg) < 1)
				wprintf(L"StartExeTransfer failed: %d, %s\n", error, berror_msg);
			else
				wprintf(L"StartExeTransfer failed: %s\n", berror_msg);
			pJob->DestroyJob(&error, &berror_msg);
			pJob->Release();
			ExitProcess(1);
		}

		// Read from the local file
		ReadFile(hLocalFile, buffer, num_read, &num_read, NULL);
		
		while (num_read > 0)
		{
			bound.cElements = num_read;
			v.parray = SafeArrayCreate(VT_UI1, 1, &bound);
			if (v.parray == NULL)
			{
				_tprintf(TEXT("Unable to allocate buffers.\n"));
				CloseHandle(hLocalFile);
				SysFreeString(berror_msg);
				delete buffer;
				return;
			}
			
			// Copy the data into a SAFEARRAY
			SafeArrayAccessData(v.parray, &v_buf);
			CopyMemory(v_buf, buffer, num_read);
			SafeArrayUnaccessData(v.parray);
			
			// Transfer the data to the remote systems
			hr = pJob->TransferExeData(v, &error, &berror_msg);
			if (FAILED(hr))
			{
				_tprintf(TEXT("TransferExeData failed\n"));
				PrintError(hr);
				CloseHandle(hLocalFile);
				SysFreeString(berror_msg);
				SafeArrayDestroy(v.parray);
				delete buffer;
				return;
			}
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"TransferExeData failed: %d, %s\n", error, berror_msg);
				else
					wprintf(L"TransferExeData failed: %s\n", berror_msg);
				CloseHandle(hLocalFile);
				SysFreeString(berror_msg);
				SafeArrayDestroy(v.parray);
				delete buffer;
				return;
			}
			
			// Is there a way to re-size a SafeArray?
			// For now, I just destroy and recreate it at each iteration
			SafeArrayDestroy(v.parray);
			
			num_read = TRANSFER_BUFFER_SIZE;
			ReadFile(hLocalFile, buffer, num_read, &num_read, NULL);
		}

		error = 0;
		hr = g_pJob->FinishExeTransfer(&error, &berror_msg);
		if (FAILED(hr))
		{
			_tprintf(TEXT("FinishExeTransfer failed: "));
			PrintError(hr);
			g_pJob->DestroyJob(&error, &berror_msg);
			g_pJob->Release();
			ExitProcess(1);
		}
		if (error)
		{
			if (wcslen(berror_msg) < 1)
				wprintf(L"FinishExeTransfer failed: %d, %s\n", error, berror_msg);
			else
				wprintf(L"FinishExeTransfer failed: %s\n", berror_msg);
			pJob->DestroyJob(&error, &berror_msg);
			pJob->Release();
			ExitProcess(1);
		}

		SysFreeString(berror_msg);
		CloseHandle(hLocalFile);
	}
	else
	{
		error = GetLastError();
		_tprintf(TEXT("CreateFile(OPEN_EXISTING) failed: "));
		PrintError(error);
		pJob->DestroyJob(&error, &berror_msg);
		pJob->Release();
		ExitProcess(1);
	}
	delete buffer;
}

// Function name	: PrintOptions
// Description	    : 
// Return type		: void 
void PrintOptions()
{
	_tprintf(TEXT("\n"));
	_tprintf(TEXT("Usage:\n"));
	_tprintf(TEXT("   MPIRun -np #processes [-push] [-env \"var1=val1|var2=val2...\"] executable [args ...]\n"));
	_tprintf(TEXT("   MPIRun -localonly #processes [-env \"var1=val1|var2=val2...\"] executable [args ...]\n"));
	_tprintf(TEXT("   MPIRun configfile [args ...]\n"));
	_tprintf(TEXT("\n"));
	_tprintf(TEXT("Config file format:\n"));
	_tprintf(TEXT("   >[opt push]\n"));
	_tprintf(TEXT("   >exe c:\\temp\\mpiprogram.exe\n"));
	_tprintf(TEXT("     OR \"c:\\temp\\sub directory\\mpiprogram.exe\"\n"));
	_tprintf(TEXT("     OR \\\\host\\share\\mpiprogram.exe\n"));
	_tprintf(TEXT("   >[env var1=val1|var2=val2|var3=val3...]"));
	_tprintf(TEXT("   >[args arg1 arg2 ...]\n"));
	_tprintf(TEXT("   >hosts\n"));
	_tprintf(TEXT("   >hostname1 #procs\n"));
	_tprintf(TEXT("   >hostname2 #procs\n"));
	_tprintf(TEXT("   >hostname3 #procs\n"));
	_tprintf(TEXT("   >...\n"));
	_tprintf(TEXT("\n"));
	_tprintf(TEXT("bracketed lines are optional\n"));
	_tprintf(TEXT("\n"));
}

struct HostNodeW
{
	WCHAR wHost[100];
	HostNodeW *next;
};

// Function name	: GetHostsFromRegistry
// Description	    : 
// Return type		: bool 
// Argument         : HostNodeW **list
bool GetHostsFromRegistry(HostNodeW **list)
{
	DWORD ret_val;
	HKEY hKey;

	// Open the MPICH root key
	if ((ret_val = RegOpenKeyExW(
			HKEY_LOCAL_MACHINE, 
			LMPICHKEY,
			0, KEY_ALL_ACCESS, &hKey)) != ERROR_SUCCESS)
	{
		return false;
	}

	// Read the hosts entry
	WCHAR pszHosts[1024];
	DWORD type, num_bytes=1024*sizeof(WCHAR);
	ret_val = RegQueryValueExW(hKey, L"Hosts", 0, &type, (BYTE *)pszHosts, &num_bytes);
	RegCloseKey(hKey);
	if (ret_val != ERROR_SUCCESS)
		return false;

	WCHAR *token = NULL;
	token = wcstok(pszHosts, L"|");
	if (token != NULL)
	{
		HostNodeW *n, *l = new HostNodeW;
		l->next = NULL;
		wcscpy(l->wHost, token);
		n = l;
		while ((token = wcstok(NULL, L"|")) != NULL)
		{
			n->next = new HostNodeW;
			n = n->next;
			n->next = NULL;
			wcscpy(n->wHost, token);
		}
		*list = l;
		return true;
	}

	return false;
}

// Function name	: GetAvailableHosts
// Description	    : This function requires g_nHosts to have been previously set.
// Return type		: void 
void GetAvailableHosts()
{
	DWORD size = 100;
	GetComputerName(g_pszFirstHost, &size);
	g_nFirstSMPProcs = 1;
	if (g_nHosts > 1)
	{
		SAFEARRAYBOUND bound;
		BSTR *bptr;
		unsigned int *pInt;
		HostNodeW *list;
		bound.cElements = g_nHosts-1;
		bound.lLbound = 0;
		
		g_vHosts.vt = VT_BSTR | VT_ARRAY;
		g_vHosts.parray = SafeArrayCreate(VT_BSTR, 1, &bound);
		g_vSMPInfo.vt = VT_I4 | VT_ARRAY;
		g_vSMPInfo.parray = SafeArrayCreate(VT_I4, 1, &bound);

		if (GetHostsFromRegistry(&list))
		{
			HostNodeW *n = list;
			for (long index=0; index<(long)bound.cElements; index++)
			{
				SafeArrayPtrOfIndex(g_vHosts.parray, &index, (void**)&bptr);
				*bptr = SysAllocString(n->wHost);
				SafeArrayPtrOfIndex(g_vSMPInfo.parray, &index, (void**)&pInt);
				*pInt = 1;
				n = n->next;
				if (n == NULL)
					n = list;
			}
			n = list;
			while (n != NULL)
			{
				list = n;
				n = n->next;
				delete list;
			}
		}
		else
		{
			printf("No hosts in the registry, processes will launch locally.\n");
			// If there are no entries in the registry then just repeat the current host name
			WCHAR whost[100];
			
			size = 100;
			GetComputerNameW(whost, &size);
			for (long index=0; index<(long)bound.cElements; index++)
			{
				SafeArrayPtrOfIndex(g_vHosts.parray, &index, (void**)&bptr);
				*bptr = SysAllocString(whost);
				SafeArrayPtrOfIndex(g_vSMPInfo.parray, &index, (void**)&pInt);
				*pInt = 1;
			}
		}
	}
	else
	{
		VariantInit(&g_vHosts);
		VariantInit(&g_vSMPInfo);
	}
}

struct HostNode
{
	TCHAR host[100];
	long nSMPProcs;
	HostNode *next;
};

// Function name	: ParseConfigFile
// Description	    : 
// Return type		: void 
// Argument         : LPTSTR filename
void ParseConfigFile(LPTSTR filename)
{
	FILE *fin;
	TCHAR buffer[1024] = TEXT("");

	VariantInit(&g_vHosts);
	VariantInit(&g_vSMPInfo);
	fin = _tfopen(filename, TEXT("r"));
	if (fin == NULL)
	{
		_tprintf(TEXT("Unable to open file: %s\n"), filename);
		ExitProcess(1);
	}

	while (_fgetts(buffer, 1024, fin))
	{
		// Check for mpi options
		if (_tcsnicmp(buffer, TEXT("mpiopt "), 7) == 0)
		{
			TCHAR *pChar = &buffer[7];
			while (_istspace(*pChar))
				pChar++;
			if (_tcsnicmp(pChar, TEXT("push"), 4) == 0)
				g_bPush = true;
		}
		else
		if (_tcsnicmp(buffer, TEXT("opt "), 4) == 0)
		{
			TCHAR *pChar = &buffer[4];
			while (_istspace(*pChar))
				pChar++;
			if (_tcsnicmp(pChar, TEXT("push"), 4) == 0)
				g_bPush = true;
		}
		else
		// Check for the name of the executable
		if (_tcsnicmp(buffer, TEXT("exe "), 4) == 0)
		{
			TCHAR *pChar = &buffer[4];
			while (_istspace(*pChar))
				pChar++;
			_tcscpy(g_pszExe, pChar);
			pChar = &g_pszExe[_tcslen(g_pszExe)-1];
			while (_istspace(*pChar) && (pChar >= g_pszExe))
			{
				*pChar = '\0';
				pChar--;
			}
		}
		else
		// Check for program arguments
		if (_tcsnicmp(buffer, TEXT("args "), 5) == 0)
		{
			TCHAR *pChar = &buffer[5];
			while (_istspace(*pChar))
				pChar++;
			_tcscpy(g_pszArgs, pChar);
			pChar = &g_pszArgs[_tcslen(g_pszArgs)-1];
			while (_istspace(*pChar) && (pChar >= g_pszArgs))
			{
				*pChar = '\0';
				pChar--;
			}
		}
		else
		// Check for environment variables
		if (_tcsnicmp(buffer, TEXT("env "), 4) == 0)
		{
			TCHAR *pChar = &buffer[4];
			while (_istspace(*pChar))
				pChar++;
			_tcscpy(g_pszEnv, pChar);
			pChar = &g_pszEnv[_tcslen(g_pszEnv)-1];
			while (_istspace(*pChar) && (pChar >= g_pszEnv))
			{
				*pChar = '\0';
				pChar--;
			}
		}
		else
		// Check for hosts
		if (_tcsnicmp(buffer, TEXT("hosts"), 5) == 0)
		{
			TCHAR *pChar;
			LPTSTR token = NULL;
			while (true)
			{
				if (_fgetts(buffer, 1024, fin) == NULL)
				{
					_tprintf(TEXT("There must be at least one host specified after the hosts tag in the config file.\n"));
					fclose(fin);
					ExitProcess(1);
				}
				token = _tcstok(buffer, TEXT(" \t\r\n"));
				if (token != NULL)
				{
					if (*token == _T('#'))
						token = NULL;
					else
					{
						_tcscpy(g_pszFirstHost, token);
						token = _tcstok(NULL, TEXT(" \t\r\n"));
						g_nFirstSMPProcs = (token == NULL) ? 1 : _ttoi(token);
						break;
					}
				}
			}
			g_nHosts = 1;

			HostNode *list = NULL, *tail = NULL;
			while (_fgetts(buffer, 1024, fin))
			{
				if (buffer[0] != _T('#'))
				{
					pChar = &buffer[_tcslen(buffer)-1];
					// Trim trailing white space
					while (_istspace(*pChar) && (pChar >= buffer))
					{
						*pChar = '\0';
						pChar--;
					}
					// If there is anything left on the line, consider it a host name
					if (_tcslen(buffer) > 0)
					{
						if (list == NULL)
						{
							tail = list = new HostNode;
							list->next = NULL;
							LPTSTR token = _tcstok(buffer, TEXT(" \t\r\n"));
							if (token != NULL)
								_tcscpy(list->host, token);
							token = _tcstok(NULL, TEXT(" \t\r\n"));
							list->nSMPProcs = (token == NULL) ? 1 : _ttoi(token);
						}
						else
						{
							tail->next = new HostNode;
							tail = tail->next;
							tail->next = NULL;
							LPTSTR token = _tcstok(buffer, TEXT(" \t\r\n"));
							if (token != NULL)
								_tcscpy(tail->host, token);
							token = _tcstok(NULL, TEXT(" \t\r\n"));
							tail->nSMPProcs = (token == NULL) ? 1 : _ttoi(token);
						}
						g_nHosts++;
					}
				}
			}
			fclose(fin);

			WCHAR wHost[100];
			SAFEARRAYBOUND bound;
			long index = 0;
			HRESULT hr;
			BSTR *pBstr = NULL;
			long *pInt = NULL;

			bound.lLbound = 0;
			bound.cElements = g_nHosts - 1;
			g_vHosts.vt = VT_ARRAY | VT_BSTR;
			g_vHosts.parray = SafeArrayCreate(VT_BSTR, 1, &bound);
			g_vSMPInfo.vt = VT_ARRAY | VT_I4;
			g_vSMPInfo.parray = SafeArrayCreate(VT_I4, 1, &bound);
			while (list)
			{
#ifdef UNICODE
				wcscpy(wHost, list->host);
#else
				mbstowcs(wHost, list->host, strlen(list->host)+1);
#endif
				hr = SafeArrayPtrOfIndex(g_vHosts.parray, &index, (void**)&pBstr);
				if (FAILED(hr))
				{
					_tprintf(TEXT("SafeArrayPtrOfIndex failed\n"));
					fclose(fin);
					ExitProcess(1);
				}
				*pBstr = SysAllocString(wHost);
				hr = SafeArrayPtrOfIndex(g_vSMPInfo.parray, &index, (void**)&pInt);
				if (FAILED(hr))
				{
					_tprintf(TEXT("SafeArrayPtrOfIndex failed\n"));
					fclose(fin);
					ExitProcess(1);
				}
				*pInt = list->nSMPProcs;
				index++;
				tail = list;
				list = list->next;
				delete tail;
			}
			return;
		}
	}
	fclose(fin);
}

// Function name	: GetAccountAndPassword
// Description	    : Attempts to read the password from the registry, 
//	                  upon failure it requests the user to provide one
// Return type		: void 
void GetAccountAndPassword()
{
	if (!ReadPasswordFromRegistry(g_pszAccount, g_pszPassword))
	{
		TCHAR ch=0;
		int index = 0;
		
		do
		{
			_ftprintf(stderr, TEXT("account: "));
			fflush(stderr);
			_getts(g_pszAccount);
		} while (_tcslen(g_pszAccount) == 0);
		
		_ftprintf(stderr, TEXT("password: "));
		fflush(stderr);

		HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
		DWORD dwMode;
		if (!GetConsoleMode(hStdin, &dwMode))
			dwMode = ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT | ENABLE_MOUSE_INPUT;
		SetConsoleMode(hStdin, dwMode & ~ENABLE_ECHO_INPUT);
		_getts(g_pszPassword);
		SetConsoleMode(hStdin, dwMode);
		
		_ftprintf(stderr, TEXT("\n"));
	}
}

// Function name	: CtrlHandlerRoutine
// Description	    : 
// Return type		: BOOL WINAPI 
// Argument         : DWORD dwCtrlType
BOOL WINAPI CtrlHandlerRoutine(DWORD dwCtrlType)
{
	fprintf(stderr, "User break\n");
	if (g_pJob != NULL)
	{
		BSTR berror_msg = SysAllocString(L"");
		LONG error;
		fprintf(stderr, "Aborting job\n");
		g_pJob->Abort(&error, &berror_msg);
		SysFreeString(berror_msg);
		g_pJob->Release();
	}
	ExitProcess(1);
	return TRUE;
}

struct RedirectInputThreadArg
{
	HANDLE hEvent;
	IStream **ppStream;
};

char g_pBuffer[1024];
HANDLE g_hBufferEvent1 = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hBufferEvent2 = CreateEvent(NULL, TRUE, FALSE, NULL);
DWORD g_num_read = 0;

// Function name	: ReadStdinThread
// Description	    : 
// Return type		: void 
void ReadStdinThread()
{
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
	while (true)
	{
		if (!ReadFile(hStdin, g_pBuffer, 1024, &g_num_read, NULL))
			return;
		if (g_num_read == 0)
			return;
		ResetEvent(g_hBufferEvent2);
		SetEvent(g_hBufferEvent1);
		WaitForSingleObject(g_hBufferEvent2, INFINITE);
	}
}

// Function name	: RedirectInputThread
// Description	    : 
// Return type		: void 
// Argument         : RedirectInputThreadArg *arg
void RedirectInputThread(RedirectInputThreadArg *arg)
{
	IMPIJob *pJob=NULL;
	HRESULT hr=S_OK;
	HANDLE hObject[2];
	long error=0;
	BSTR berror_msg;
	DWORD ret_val;

	hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

	berror_msg = SysAllocString(L"");
	CoGetInterfaceAndReleaseStream (*arg->ppStream, IID_IMPIJob, (void**) &pJob);
	delete arg->ppStream;

	DWORD dwThreadID;
	HANDLE hRSIThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadStdinThread, NULL, 0, &dwThreadID);

	hObject[0] = arg->hEvent;
	hObject[1] = g_hBufferEvent1;

	berror_msg = SysAllocString(L"");
	while (true)
	{
		ret_val = WaitForMultipleObjects(2, hObject, FALSE, INFINITE);
		if (ret_val == WAIT_OBJECT_0+1)
		{
			if (g_num_read > 0)
			{
				SAFEARRAYBOUND bound;
				VARIANT vInput;
				void *pBuf;

				VariantInit(&vInput);
				bound.lLbound = 0;
				bound.cElements = g_num_read;
				vInput.vt = VT_UI1 | VT_ARRAY;
				vInput.parray = SafeArrayCreate(VT_UI1, 1, &bound);

				SafeArrayAccessData(vInput.parray, &pBuf);
				memcpy(pBuf, g_pBuffer, g_num_read);
				SafeArrayUnaccessData(vInput.parray);

				error = 0;
				hr = pJob->PutJobInput(vInput, 0, &error, &berror_msg);
				if (FAILED(hr))
				{
					VariantClear(&vInput);
					printf("PutJobInput failed: %d\n", hr);
					PrintError(hr);
					break;
				}
				if (error)
				{
					VariantClear(&vInput);
					if (wcslen(berror_msg) < 1)
						wprintf(L"PutJobInput failed: %d %s\n", error, berror_msg);
					else
						wprintf(L"PutJobInput failed: %s\n", berror_msg);
					break;
				}
				VariantClear(&vInput);
			}
			ResetEvent(g_hBufferEvent1);
			SetEvent(g_hBufferEvent2);
		}
		else
		{
			//printf("g_hFinishedEvent signalled\n");
			break;
		}
	}

	pJob->Release();
	CoUninitialize();
}

// Function name	: main
// Description	    : 
// Return type		: void 
// Argument         : int argc
// Argument         : TCHAR *argv[]
void main(int argc, TCHAR *argv[])
{
	SetConsoleCtrlHandler(CtrlHandlerRoutine, TRUE);
	long launch_flags = 0;

	if (argc == 1)
	{
		PrintOptions();
		return;
	}

	if (GetOpt(argc, argv, "-nompi"))
	{
		g_bNoMPI = true;
		launch_flags |= MPI_LAUNCH_OPT_NOMPI;
	}
	g_bPush = GetOpt(argc, argv, "-push");
	GetOpt(argc, argv, "-env", g_pszEnv);

	DWORD dwType;
	if (GetBinaryType(argv[1], &dwType))
	{
		// The first argument is an executable so set things up to run one process
		DWORD size = 100;

		g_nHosts = 1;
		_tcscpy(g_pszExe, argv[1]);
		_tcscpy(g_pszArgs, TEXT(""));
		for (int i=2; i<argc; i++)
		{
			_tcscat(g_pszArgs, argv[i]);
			if (i < argc-1)
				_tcscat(g_pszArgs, TEXT(" "));
		}
		g_bPush = false;
		VariantInit(&g_vHosts);
		GetComputerName(g_pszFirstHost, &size);
	}
	else
	{
		if (GetOpt(argc, argv, "-np", &g_nHosts))
		{
			if (g_nHosts < 1)
			{
				printf("Error: must specify a number greater than 0 after the -np option\n");
				return;
			}
			if (argc < 2)
			{
				printf("Error: not enough arguments.\n");
				return;
			}
			_tcscpy(g_pszExe, argv[1]);
			g_pszArgs[0] = TEXT('\0');
			for (int i=2; i<argc; i++)
			{
				_tcscat(g_pszArgs, argv[i]);
				if (i < argc-1)
					_tcscat(g_pszArgs, TEXT(" "));
			}
			GetAvailableHosts();
		}
		else
		if (GetOpt(argc, argv, "-localonly", &g_nHosts))
		{
			bool bDoSMP = !GetOpt(argc, argv, "-tcp");
			if (g_nHosts < 1)
			{
				printf("Error: must specify a number greater than 0 after the -localonly option\n");
				return;
			}
			if (argc < 2)
			{
				printf("Error: not enough arguments.\n");
				return;
			}
			TCHAR pszTempExe[MAX_PATH], *namepart;
			_tcscpy(g_pszExe, argv[1]);
			GetFullPathName(g_pszExe, MAX_PATH, pszTempExe, &namepart);
			// Quote the executable in case there are spaces in the path
			_stprintf(g_pszExe, TEXT("\"%s\""), pszTempExe);
			g_pszArgs[0] = TEXT('\0');
			for (int i=2; i<argc; i++)
			{
				_tcscat(g_pszArgs, argv[i]);
				if (i < argc-1)
					_tcscat(g_pszArgs, TEXT(" "));
			}
			//* Don't use the launcher
			RunLocal(bDoSMP);
			return;
			/*/ Use the launcher
			// Create a list with the local hostname repeated for each entry
			SAFEARRAYBOUND bound;
			BSTR *bptr;
			long *pLong;
			WCHAR whost[100];
			DWORD size = 100;
			bound.cElements = g_nHosts-1;
			bound.lLbound = 0;
			
			GetComputerName(g_pszFirstHost, &size);

			if (bDoSMP)
			{
				VariantInit(&g_vHosts);
				VariantInit(&f_vSMPInfo);

				g_nFirstSMPProcs = g_nHosts;
			}
			else
			{
				g_nFirstSMPProcs = 1;

				g_vHosts.vt = VT_BSTR | VT_ARRAY;
				g_vHosts.parray = SafeArrayCreate(VT_BSTR, 1, &bound);
				g_vSMPInfo.vt = VT_I4 | VT_ARRAY;
				g_vSMPInfo.parray = SafeArrayCreate(VT_I4, 1, &bound);
			
				size = 100;
				GetComputerNameW(whost, &size);
				for (long index=0; index<(long)bound.cElements; index++)
				{
					SafeArrayPtrOfIndex(g_vHosts.parray, &index, (void**)&bptr);
					*bptr = SysAllocString(whost);
					SafeArrayPtrOfIndex(g_vSMPInfo.parray, &index, (void**)&pLong);
					*pLong = 1;
				}
			}
			//*/
		}
		else
		{
			ParseConfigFile(argv[1]);
			if ((_tcslen(g_pszArgs) > 0) && (argc > 2))
				_tcscat(g_pszArgs, TEXT(" "));
			for (int i=2; i<argc; i++)
			{
				_tcscat(g_pszArgs, argv[i]);
				if (i < argc-1)
					_tcscat(g_pszArgs, TEXT(" "));
			}
		}
	}

	GetAccountAndPassword();

	TCHAR pszTempExe[MAX_PATH], *namepart;
	GetFullPathName(g_pszExe, MAX_PATH, pszTempExe, &namepart);
	// Quote the executable in case there are spaces in the path
	_stprintf(g_pszExe, TEXT("\"%s\""), pszTempExe);

	TCHAR pszDir[MAX_PATH] = TEXT(".");
	GetCurrentDirectory(MAX_PATH, pszDir);

	LONG error = 0;
	BSTR berror_msg;
	BSTR bCommonArgs, bAccount, bPassword, bExe, bNic, bDir, bEnv;
	HRESULT hr = S_OK;

	berror_msg = SysAllocString(L"");
#ifdef UNICODE
	bExe = SysAllocString(g_pszExe);
	bCommonArgs = SysAllocString(g_pszArgs);
	bAccount = SysAllocString(g_pszAccount);
	bPassword = SysAllocString(g_pszPassword);
	bNic = SysAllocString(g_pszFirstHost);
	bDir = SysAllocString(pszDir);
	bEnv = SysAllocString(g_pszEnv);
#else
	WCHAR wTemp[MAX_PATH];
	mbstowcs(wTemp, g_pszExe, strlen(g_pszExe)+1);
	bExe = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszAccount, strlen(g_pszAccount)+1);
	bAccount = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszPassword, strlen(g_pszPassword)+1);
	bPassword = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszArgs, strlen(g_pszArgs)+1);
	bCommonArgs = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszFirstHost, strlen(g_pszFirstHost)+1);
	bNic = SysAllocString(wTemp);
	mbstowcs(wTemp, pszDir, strlen(pszDir)+1);
	bDir = SysAllocString(wTemp);
	mbstowcs(wTemp, g_pszEnv, strlen(g_pszEnv)+1);
	bEnv = SysAllocString(wTemp);
#endif

	fprintf(stderr, "Connecting to %s\n", g_pszFirstHost);fflush(stderr);
	//wprintf(L"bNic: %s\n", bNic);fflush(stdout);
	if (Connect(g_pszFirstHost, g_pJob))
	{
		fprintf(stderr, "Creating MPICH Job\n");fflush(stderr);
		error = 0;
		hr = g_pJob->CreateJob(g_vHosts, g_vSMPInfo, g_nFirstSMPProcs, bNic, &error, &berror_msg);

		if (SUCCEEDED(hr))
		{
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"Unable to create the job: Error(%d)", error);
				else
					wprintf(L"Unable to create the job: %s", berror_msg);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				SysFreeString(bEnv);
				return;
			}

			if (g_bPush)
			{
				fprintf(stderr, "Transfering executable to remote systems\n");fflush(stderr);
				TransferFile(g_pJob, pszTempExe);
			}
			else
			{
				error = 0;
				hr = g_pJob->SetExecutable(bExe, &error, &berror_msg);
				if (FAILED(hr))
				{
					wprintf(L"SetExecutable failed: ");
					PrintError(hr);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					SysFreeString(bEnv);
					return;
				}
				if (error)
				{
					if (wcslen(berror_msg) < 1)
						wprintf(L"SetExecutable failed: Error %d\n", error);
					else
						wprintf(L"SetExecutable failed: %s\n", berror_msg);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					SysFreeString(bEnv);
					return;
				}
			}

			error = 0;
			hr = g_pJob->SetWorkingDirectory(bDir, &error, &berror_msg);
			if (FAILED(hr))
			{
				_tprintf(TEXT("SetWorkingDirectory failed: "));
				PrintError(hr);
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				SysFreeString(bEnv);
				return;
			}
			if (error)
			{
				if (wcslen(berror_msg) < 1)
					wprintf(L"SetWorkingDirectory failed: Error %d\n", error);
				else
					wprintf(L"SetWorkingDirectory failed: %s\n", berror_msg);
				g_pJob->DestroyJob(&error, &berror_msg);
				g_pJob->Release();
				CoUninitialize();
				SysFreeString(berror_msg);
				SysFreeString(bExe);
				SysFreeString(bCommonArgs);
				SysFreeString(bAccount);
				SysFreeString(bPassword);
				SysFreeString(bNic);
				SysFreeString(bDir);
				SysFreeString(bEnv);
				return;
			}
			
			if (wcslen(bEnv) > 0)
			{
				//fprintf(stderr, "Setting environment variables\n");fflush(stderr);
				hr = g_pJob->SetEnvironmentVariables(bEnv, &error, &berror_msg);
				if (FAILED(hr))
				{
					_tprintf(TEXT("SetEnvironmentVariables failed: "));
					PrintError(hr);
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					SysFreeString(bEnv);
					return;
				}
				if (error)
				{
					if (wcslen(berror_msg) < 1)
						wprintf(L"SetEnvironmentVariables failed: Error %d\n", error);
					else
						wprintf(L"SetEnvironmentVariables failed: %s\n", berror_msg);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					SysFreeString(bEnv);
					return;
				}
			}

			error = 0;
			fprintf(stderr, "Launching job\n");fflush(stderr);
			hr = g_pJob->LaunchJob(bCommonArgs, bAccount, bPassword, launch_flags, &error, &berror_msg);
			
			if (SUCCEEDED(hr))
			{
				if (error)
				{
					if (wcslen(berror_msg) < 1)
						wprintf(L"LaunchJob failed: Error %d\n", error);
					else
						wprintf(L"LaunchJob failed: %s\n", berror_msg);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					return;
				}

				RedirectInputThreadArg *pArg = new RedirectInputThreadArg;
				pArg->hEvent = g_hFinishedEvent;
				pArg->ppStream = new IStream*;
				// Marshall pJob to a thread which redirects user input
				hr = CoMarshalInterThreadInterfaceInStream(IID_IMPIJob, g_pJob, pArg->ppStream);
				if (FAILED(hr))
				{
					g_pJob->Abort(&error, &berror_msg);
					g_pJob->Release();
					SysFreeString(berror_msg);
					printf("CoMarshalInterThreadInterfaceInStream failed.\n");
					CoUninitialize();
					delete pArg;
					return;
				}

				DWORD dwThreadID;
				HANDLE hRedirectInputThread = CreateThread(
					NULL, 0, 
					(LPTHREAD_START_ROUTINE)RedirectInputThread,
					pArg, 
					0, &dwThreadID);
				
				if (hRedirectInputThread == NULL)
				{
					error = GetLastError();
					printf("CreateThread failed: error %d\n", error);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					SysFreeString(berror_msg);
					CoUninitialize();
					delete pArg;
					return;
				}

				HANDLE hStdout;
				hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
				if (hStdout == INVALID_HANDLE_VALUE)
				{
					error = GetLastError();
					_tprintf(TEXT("GetStdHandle failed: Error %d\n"), error);
					g_pJob->DestroyJob(&error, &berror_msg);
					g_pJob->Release();
					CoUninitialize();
					SysFreeString(berror_msg);
					SysFreeString(bExe);
					SysFreeString(bCommonArgs);
					SysFreeString(bAccount);
					SysFreeString(bPassword);
					SysFreeString(bNic);
					SysFreeString(bDir);
					return;
				}

#ifdef MULTI_COLOR_OUTPUT
				WORD g_ConsoleAttribute;
				CONSOLE_SCREEN_BUFFER_INFO info;
				// Save the state of the console so it can be restored after each change
				GetConsoleScreenBufferInfo(hStdout, &info);
				g_ConsoleAttribute = info.wAttributes;
#endif
				long more = 1, from = 0;
				VARIANT *v = new VARIANT;
				DWORD num_written, num_elements;
				void *pBuf;
				
				v->vt = VT_UI1 | VT_ARRAY;
				SAFEARRAYBOUND bound;
				bound.lLbound = 0;
				bound.cElements = 0;
				v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
				
				while (more == 1)
				{
					error = 0;
					hr = g_pJob->GetJobOutput(v, &from, &more, &error, &berror_msg);
					if (FAILED(hr))
					{
						_tprintf(TEXT("DCOM failure: GetJobOutput()\n"));
						PrintError(hr);
						SysFreeString(berror_msg);
						SysFreeString(bExe);
						SysFreeString(bCommonArgs);
						SysFreeString(bAccount);
						SysFreeString(bPassword);
						SysFreeString(bNic);
						SysFreeString(bDir);
						VariantClear(v);
						delete v;
						return;
					}
					if (error)
					{
						if (wcslen(berror_msg) < 1)
							wprintf(L"Error %d", error);
						else
							wprintf(L"%s", berror_msg);
						SysFreeString(berror_msg);
						SysFreeString(bExe);
						SysFreeString(bCommonArgs);
						SysFreeString(bAccount);
						SysFreeString(bPassword);
						SysFreeString(bNic);
						SysFreeString(bDir);
						VariantClear(v);
						delete v;
						return;
					}
					if (v->parray != NULL)
					{
						num_elements = v->parray->rgsabound->cElements;
						if (num_elements > 0)
						{
							SafeArrayAccessData(v->parray, &pBuf);
#ifdef MULTI_COLOR_OUTPUT
							SetConsoleTextAttribute(hStdout, aConsoleColorAttribute[from%NUM_OUTPUT_COLORS]);
							WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
							SetConsoleTextAttribute(hStdout, g_ConsoleAttribute);
#else
							WriteFile(hStdout, pBuf, num_elements, &num_written, NULL);
#endif
							FlushFileBuffers(hStdout);
							SafeArrayUnaccessData(v->parray);
							// Destroy the array.
							SafeArrayDestroy(v->parray);
							// Create a new and empty array.
							v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
						}
						else
						{
							SafeArrayDestroy(v->parray);
							v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
						}
					}
					else
					{
						v->parray = SafeArrayCreate(VT_UI1, 1, &bound);
					}
				}
				
				SafeArrayDestroy(v->parray);
				delete v;
				
				//_tprintf(TEXT("\n"));
#ifdef MULTI_COLOR_OUTPUT
				SetConsoleTextAttribute(hStdout, g_ConsoleAttribute);
#endif

				// Signal the redirect input thread to terminate
				SetEvent(g_hFinishedEvent);
				WaitForSingleObject(hRedirectInputThread, 1000);
				CloseHandle(hRedirectInputThread);
			}
			else
			{
				_tprintf(TEXT("LaunchJob failed: "));
				PrintError(hr);
				fflush(stdout);
			}
			
			g_pJob->DestroyJob(&error, &berror_msg);
			g_pJob->Release();
		}
		else
		{
			_tprintf(TEXT("Unable to create a job: "));
			PrintError(hr);
		}

		CoUninitialize();
	}
	else
	{
		_tprintf(TEXT("Unable to connect to '%s'\n"), g_pszFirstHost);
	}

	SysFreeString(berror_msg);
	SysFreeString(bExe);
	SysFreeString(bCommonArgs);
	SysFreeString(bAccount);
	SysFreeString(bPassword);
	SysFreeString(bNic);
	SysFreeString(bDir);
	SysFreeString(bEnv);
}
