#ifndef NT_GLOBAL_CPP_H
#define NT_GLOBAL_CPP_H

#include <winsock2.h>
#include <windows.h>

#include "nt_common.h"
#include "nt_tcp_sockets.h"
#include "nt_global.h"
#include "nt_log.h"
#include "MessageQueue.h"
#include "Database.h"

#define MULTI_COLOR_OUTPUT

#define NT_TCP_CONNECT_INFO_CMD		0
#define NT_TCP_PROCESS_INFO_CMD		1
#define NT_TCP_END_CMD				2

#define NT_TCP_CTRL_CMD_INIT_DATA_TO_ROOT		0
#define NT_TCP_CTRL_CMD_PROCESS_CONNECT_INFO	1
#define NT_TCP_CTRL_CMD_PROCESS_INFO			2
#define NT_TCP_CTRL_CMD_POST_IN_DONE			3
#define NT_TCP_CTRL_CMD_ALL_IN_DONE				4
#define NT_TCP_CTRL_CMD_ABORT					5

#define NT_COMM_CMD_ADD_SOCKET		1
#define NT_COMM_CMD_EXIT			2

#define NT_HOSTNAME_LEN				100
#define NT_EXENAME_LEN				256
#define NT_TCP_WAIT_TIME			10000

#define NT_MSG_READING_TAG			0
#define NT_MSG_READING_LENGTH		1
#define NT_MSG_READING_BUFFER		2

// global structures
struct NT_Tcp_shm_ProcEntry
{
	SOCKET sock;					// Communication socket
	WSAEVENT sock_event;			// Communication socket event
	struct Message					// Structure accessed by completion port or via thread to store the current message
	{
		int tag;
		int length;
		void *buffer;
		int nRemaining;
		DWORD nRead;
		OVERLAPPED ovl;
		MessageQueue::MsgQueueElement *pElement;
		int state;					// NT_MSG_READING_TAG, NT_MSG_READING_LENGTH, NT_MSG_READING_BUFFER
	} msg;

	int shm;						// FALSE(0) or TRUE(1) if this host can be reached through shared memory
	int via;						// FALSE(0) or TRUE(1) if this host can be reached through VI

	int listen_port;				// Port where thread is listening for connections
	int control_port;				// Port where thread is listening for control message connections

	// Description of process
	long pid;						// process id
	char host[NT_HOSTNAME_LEN];		// host where process resides
	char exename[NT_EXENAME_LEN];	// command line launched on the node

	HANDLE hValidDataEvent;			// Event signalling the data in this structure is valid 
									// This does not include sock and sock_event
};

struct ControlLoopClientArg
{
	SOCKET sock;
	HANDLE sock_event;
};

// nt_tcp_shm_comport.cpp
extern HANDLE g_hCommPortThread;
extern HANDLE g_hCommPort;
extern HANDLE g_hCommPortEvent;
extern int g_nCommPortCommand;
extern int g_NumCommPortThreads;
int ConnectTo(int remote_iproc);
void CommPortThread(HANDLE hReadyEvent);

// nt_tcp_shm_priv.cpp
extern NT_Tcp_shm_ProcEntry *g_pProcTable;
extern bool g_bInNT_Tcp_shm_End;
extern MessageQueue g_MsgQueue;
extern char g_pszHostName[NT_HOSTNAME_LEN];
extern char g_pszRootHostName[NT_HOSTNAME_LEN];
extern int g_nRootPort;
extern HANDLE *g_aRunningProcess;
extern char g_pszJobID[100];
extern bool g_bUseDatabase;
extern Database g_Database;

// nt_smp.cpp
extern int g_nSMPLow;
extern int g_nSMPHigh;
extern HANDLE *g_hShpSendCompleteEvent;
extern HANDLE *g_hProcesses;
void InitSMP();
void EndSMP();
void NT_ShmSend(int type, void *buffer, int length, int to);
void GetShmemClick();

// nt_tcp_shm_control_loop.cpp
void ControlLoopClientThread(ControlLoopClientArg *arg);
void ControlLoopThread(HANDLE hReadyEvent);
bool SendInitDataToRoot();
bool GetProcessConnectInfo(int iproc);
bool GetProcessInfo(int iproc);
bool SendInDoneMsg();
extern HANDLE g_hAllInDoneEvent;
extern HANDLE g_hOkToPassThroughDone;
extern HANDLE g_hControlLoopThread;
extern HANDLE g_hStopControlLoopEvent;
extern HANDLE g_hEveryoneConnectedEvent;

#endif
